#ifndef lint
static  char sccsid[] = "@(#)pr.c 1.6 91/05/07 Copyr 1990 SMI";
#endif

/*
 * Copyright (c) 1991, Legato Systems Incorporated.
 *
 * All rights reserved.
 */

/*
 * The ``presto'' device driver.  This driver implements a disk driver
 * using non-volatile memory as a cache.  This benefits NFS server write
 * performance since all NFS writes are required to be synchronous to
 * permanent storage.  For presto, these writes will be synchronous to
 * our non-volatile memory.  This reduces the disk arm bottleneck present
 * on most servers.  In fact, things like inode and indirect blocks which
 * will be flushed back to disk on every write for sequential write over
 * NFS can profit enormously because they will remain hot.  This avoids
 * 2/3's of all the disk write operations for NFS sequential writes of
 * large files.
 *
 * This driver works as a "layer" between other block drivers and the rest
 * of the UNIX kernel.  Presto's stubs replace the original driver's
 * entry points in the cdev and bdev switches.  Whenever Presto needs to
 * perform actual IO, as when the cache needs filling or draining, Presto
 * will call the "layered" driver's entry points (strategy, close, read and
 * write).
 */

#include "pr.h"

#if NPR > 0

#define COPY_PROTECT		/* verify the mm6704 board is from Legato */
/* #define INIT_COPY_PROTECT 	/* "mark" the mm6704 board as from Legato */
/* #define DSW_DEBUG		/* tell us what going on with the SMARTwatch */

#if defined(INIT_COPY_PROTECT) && ! defined(PRDEBUG)
#define PRDEBUG
#endif

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/buf.h>
#include <sys/map.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/kernel.h>
#include <sys/file.h>
#include <sys/vmmac.h>
#include <sys/reboot.h>
#include <sys/errno.h>

#include <machine/pte.h>
#include <machine/mmu.h>
#include <machine/seg_kmem.h>
#include <vm/hat.h>
#include <sundev/presto.h>
#include <sundev/mbvar.h>

#ifdef PRDEBUG
#include <sys/debug.h>
#else
#define ASSERT(x)
#endif

#ifndef TRUE
#define TRUE	1
#endif

#ifndef FALSE
#define FALSE	0
#endif

#ifdef sun
static caddr_t praddr;			/* address of mmap'ed pr space */
static struct nv *nvhp;			/* address of mmap'ed NV header */
static u_int *prchksums;		/* address of mmap'ed check sums */
static struct nvbd *nvbdp;		/* address of mmap'ed buffer data */
static struct ps_statclk *psaddr;	/* address of mmap'ed status register */

/*
 * All Sun's support direct virtual mapping of the bus.
 * This allows the presto driver to directly access the
 * presto board using standard routines like bcopy.
 */
#define BUSWRITE(busaddr, vaddr, count)	bcopy((vaddr), (busaddr), (count))
#define BUSREAD(busaddr, vaddr, count)	bcopy((busaddr), (vaddr), (count))
#define BUSZERO(busaddr, count)		bzero((busaddr), (count))

/*
 * BSA: Ugly kludge below is because pte mapping changed between Sun4c & Sun4m
 * This driver only plays with the cacheable bits of this struct. So,
 * to make this .o file arch independent, I'm hardcoding this union with
 * postions for both cache bits to be used in prsbusattach().
 * This routine then uses appropriate bit at run time depending on arch.
 * IMPORTANT NOTE: In the unlikely event that the pte struct definition
 * changes for the sun4c or sun4m architectures, this union also needs to
 * be updated.
 */
typedef union un_pte {
	struct pte base;
	struct ovlay {
		u_int garbag1:3;
		u_int pg_nc:1;		/* sun4c: no cache bit */
		u_int garbag2:20;
		u_int Cacheable:1;	/* sun4m: cache bit */
		u_int garbag3:7;
	} cache;
} UN_PTE;

#else
static caddr_t praddr;			/* address of mmap'ed pr space */
static struct nv *nvhp;			/* S/VMEbus memory board address */
static u_int *prchksums;		/* address of mmap'ed check sums */
static struct nvbd *nvbdp;		/* address of mmap'ed buffer data */
static struct ps_statclk *psaddr;	/* S/VMEbus status register address */
	Error - don't know how to deal with the bus on this machine!
#endif

/*
 * PRBALIGN is used for aligning mapped addresses
 * within praddr to PRBSIZE offsets within the NV memory.
 * The parameter is a pointer within the praddr space.
 */
#define PRBALIGN(x) \
	(praddr + ((((caddr_t)(x) - praddr) + PRBSIZE - 1) & ~(PRBSIZE - 1)))

static u_char *battery_registers;	/* battery enabled and ok flags */
static u_long *prdirty_led;		/* light when board has dirty data */

enum battery_modes {
	BAT_NONE 	= 0,		/* batteries are always bad */
	BAT_BAD 	= 1,		/* batteries are bad (and tell them) */
	BAT_FAKE 	= 2,		/* fakebattery is the value to use */
	BAT_PRETEND 	= 3,		/* pretend batteries (always good) */
	BAT_MM6704 	= 4,		/* check copy protect stuff in clock */
	BAT_MM1800 	= 5		/* Just The Batteries Mam */
};

static enum battery_modes  battery_mode = BAT_NONE;

static int Pr_normal;			/* set at boottime */
static int Pr_dontcare;			/* set at boottime */
static int Pr_allbattenabled;		/* set at boottime */
static int Pr_allbattlow;		/* set at boottime */

static u_long pr_batt_check_rate = 30;	/* check batteries this often */
static u_long pr_batt_check_time;	/* next time to check batteries */


struct bufhd prbufhash[PRBUFHSZ];	/* heads of hash lists */
struct diskhd prbfreelist[PRBQUEUES];	/* heads of available lists */
struct buf *prbufs;			/* head of array of presto buffers */
#define USE_B_CHAIN
#ifndef USE_B_CHAIN
int *proldflags;			/* head of array of presto oldflags */
#endif
int prnbufs;				/* number of allocated presto bufs */
struct presto_status prstatus;		/* presto status structure */
u_long pr_stats_cleared = 0;		/* time.tv_sec when stats cleared */
u_long pr_went_down = 0;		/* time.tv_sec when presto went down */
u_long pr_went_up = 0;			/* time.tv_sec when presto went up */

/*
 * BSA: Added code for NVRAM scrubbing support. Tagged on to PRCHKSUMALL
 * logic for Sun based systems. Every nth write will cause us to step 
 * linearly thru buffers checksuming next buffer in memory. 
 * This guarantees that, eventually, ALL NVRAM buffers will be checksum'd.
 * Declared statically here so they can be patched dynamically, if necessary.
 * Performance & functionality improvement because now we only check
 * 1 time in n (n=50), & also by going sequentially thru buffers, we have
 * a much better chance of eventually getting to every buffer on the board,
 * instead of just the one that's currently being written.
 */
#ifndef SUN_RELEASE
#define	SUN_RELEASE	1
#elif defined(SUN_RELEASE) && SUN_RELEASE == 0
#undef	SUN_RELEASE
#define	SUN_RELEASE	1
#endif
#ifndef OPENPROMS
#define OPENPROMS
#endif
/*
 * #define SUNSCRUB_DEBUG
 * #define SUNSCRUB_BIG_DEBUG
 */
#define SUNSCRUB
#ifdef SUNSCRUB
static int pr_ios = 0;			/* Counter for trigger at nth I/O */
#ifdef SUNSCRUB_DEBUG
static int pr_nth = 5;			/* How many consitute an nth ? */
#else
static int pr_nth = 50;
#endif
int pr_next_scrub = 0;			/* Number of next buffer to check */
#endif

extern void gets( /* char *buf */ );			/* from swapgeneric.c */
extern int machineid( /* void */);			/* from machdep.c */

/*
 * Forward declarations for some internal routines
 */
static void prreset_finish( /* void */);
static void prbwrite( /* struct buf *bp */ );
static void prbwait( /* struct buf *bp */ );
static void prbremfree( /* struct buf *bp */ );
static void prbrelse( /* struct buf *bp */ );
static void bptonvbp( /* struct buf *bp */ );
static int prflushdev( /* dev_t dev, bool_t inval */ );

#ifdef PRDEBUG
#define PRDEBUG1(s, a, b, c, d)	\
	if (prdebug >= 1) \
		printf((s), (a), (b), (c), (d))
#define PRDEBUG2(s, a, b, c, d) \
	if (prdebug >= 2) \
		printf((s), (a), (b), (c), (d))
#define PRDEBUG3(s, a, b, c, d) \
	if (prdebug >= 3) \
		printf((s), (a), (b), (c), (d))
#define PRTRACE(s, a, b) \
	if (prtrace) \
		printf((s), (a), (b))
#define PRCHECK() \
	if (prcheck) \
		prchk()

/*
 * When we are compiled for debugging, we need to insure that
 * prchk() see buffers with nicely connected hash and free chains.
 */
#define PRFIXFREE(bp)		(bp)->av_forw = (bp)->av_back = (bp)

static int prtrace = 0;		/* tracing variable */
static int prdebug = 0;		/* debugging level */
static int prcheck = 0;		/* check listed list flag */
#else
#define PRDEBUG1(s, a, b, c, d)
#define PRDEBUG2(s, a, b, c, d)
#define PRDEBUG3(s, a, b, c, d)
#define PRTRACE(s, a, b)
#define PRCHECK()
#define PRFIXFREE(bp)
#endif

static u_int prallocmem = 0;	/* if non-zero, kmem_alloc the presto mem */
static int prreinit = 0;	/* reinitialize everything at boot time */

/*
 * When a buffer is given off to the real device for a write operation,
 * it is added to the active list and is marked as B_ACTIVE until it is reaped.
 */
#define B_ACTIVE	0x80000000

static struct buf **practive;	/* list of presto buf's given to the drivers */

static int prwrite_percent = 2;	/* % of dirty to write when we are out */
static u_int prprewrite = 1;	/* if true, use new ``prewrite strategy'' */
static u_int prprewritenum = 4;	/* if non-zero, amount to flush on prewrites */
static u_int prdesnavail = 5;	/* patchable # buffers to strive for */
static u_int prdesavail;	/* calculated # of free bufs to strive for */
static u_int prcontigbufs = 0;	/* how many (if any) contig buffers to write */
static u_int prcontigspan = 5;	/* # of blocks to look forward for contigs */
#ifdef PRDEBUG
u_int prcontigswritten[10];	/* how many (if any) contig buffers written */
u_int prblkno[0x100];		/* the blkno of the last 0x100 block written */
u_int prblkno_i;		/* the index into above (always mask by 0xff) */
u_int proverlap, prdelwri, prlookedat;
#endif /* PRDEBUG  */

extern int nprbdev;		/* the max major # of a presto block device */
extern struct prtab *prtabs[];	/* a per block device information table */
extern int prflags;		/* global flags for all devices */
caddr_t prfastpool;		/* pool of indirect write staging buffers */
int pr_bounce_active;		/* count of staging buffers currently active */
int pr_bounce_max;		/* high tide mark of staging buffers */
int pr_bounce_limit = 16;	/* limit on # of bounce buffers to alloc */
int pr_write_limit = 32;	/* limit on # of active bufs to write */

#define BOUNCEIO(p, min_d) \
    ((prflags & PR_BOUNCEIO) || (isset(((p)->pt_bounceio.bits), (min_d))))

/*
 * Variables for controlling how we deal with async writes.
 * #define pr_cache_async_occasionally to be 0 and
 * #define pr_nocache_async to be 0 to completely remove this code.
 */
#if !defined(pr_cache_async_occasionally)
int pr_cache_async_occasionally = 7;	/* cache async writes this often */
#endif
#if !defined(pr_nocache_async)
int pr_nocache_async = 0;		/* if !0 don't cache this async write */
#endif

/* a macro which tests if a particular block device # is "prestoized" */
#define PR_MAJORBDEV_OK(maj_d) 	(((maj_d) >= 0) && \
				((maj_d) < nprbdev) && \
				(prtabs[(maj_d)] != 0) && \
				(prtabs[(maj_d)]->pt_strategy != NULL))

#define PRSTEP		512		/* step size when sizing memory */

static u_char batterystatus = 0;	/* last battery status value */
static u_int fakebattery = 0;		/* if !0, use this value */

#ifdef OPENPROMS
static int prmakecachable = 0;		/* make Prestoserve memory cachable? */
#endif

/*
 *	Presto Cache (some theory of operation)
 *
 *	At boot time: Presto starts in the PRDOWN state.
 *		If there are dirty buffers, we set the prreadonly flag
 *		to turn the presto cache into a read-only cache.  We also
 *		set the prresetneeded flag, so once all the buffers are
 *		flushed, the presto data structures will be reset with the
 *		current selected size.  All entry points to the presto driver
 *		check prresetneeded, and if set they will call prreset(), which
 *		will attempt to flush the presto cache.  Until prresetneeded is
 *		cleared, presto must search the presto cache for all io
 *		requests on presto-ized drivers.  Presto will end in the PRDOWN
 *		state after successfully flushing all dirty buffers found after
 *		a reboot.
 *			boot: PRDOWN [-> PRERROR] -> PRDOWN
 *		Thus, when prresetneeded is set, *any* partition may have data
 *		in the Presto cache, regardless of that partition's enabled
 *		state.
 *
 *	After boot time (i.e. after prresetneeded is clear):
 *		a device may have data in the presto cache:
 *			if and only if the device is enabled.
 *			if and only if presto is not DOWN.
 *		presto may be taken DOWN:
 *			if and only if all dirty data are successfully flushed.
 *		a device partition may be disabled:
 *			if and only if all dirty data for that partition are
 *			successfully flushed.
 *		new write data may be cached for a partition:
 *			if and only if prreadonly is not set.
 *		if presto is in the PRERROR state
 *			then prreadonly must be set.
 */

/*
 * We have two types of global flushes.  Flushes on a particular partition,
 * and flushes of the entire presto write cache.  When flushing a specific
 * device, we do not allow that device to be PRDISABLED until the flushing
 * results in the error free transfer of all intercepted writes for that device.
 * When flushing the entire presto write cache, we set prreadonly, which
 * turns the presto cache into a read-only cache, until all dirty data is
 * flushed.
 *
 * Someday we might want to convert from a read-only system which effects all
 * presto partitions, to a per partition, or maybe per-device, read-only system.
 */
static int prresetneeded;		/* perform prreset with first context */
static int prresetting;			/* prreset has started */
static int prreadonly;			/* set when flushing all buffers */

/*
 * We use two different ``devices'' for auto configuration -
 * one for the memory and one for the status and clock registers
 * because they are at different selectable addresses on the bus
 * and the auto configuration code can only deal with one address
 * per device.
 *
 * The `pr' device is the handle on doing ioctl to control presto
 * (normally opened as /dev/pr0 by /usr/etc/presto program).
 * If the presto memory board cannot be found, /dev/pr0 cannot
 * be opened and thus presto cannot be enabled.
 */

/*
 * Driver information for auto-configuration stuff.
 */
int propen(), prioctl(), prmmap();
extern int nodev(), nulldev();

#ifdef OPENPROMS
static int prsbusidentify(), prsbusattach();

struct dev_ops pr_ops = {
	1,			/* dev_ops revision number */
	prsbusidentify,		/* sbus confirm device ID */
	prsbusattach,		/* sbus attach routine */
	propen,			/* open */
	nulldev,		/* close */
	nodev,			/* read */
	nodev,			/* write */
	nodev,			/* strategy */
	nodev,			/* dump */
	nodev,			/* psize */
	prioctl,		/* ioctl */
	nulldev,		/* reset */
	prmmap			/* mmap */
};

extern addr_t map_regs( /* addr_t addr, u_int size, int space */ );
extern void unmap_regs( /* addr_t addr, u_int size */ );
#endif /* OPENPROMS */

/*
 * The presto `pr' auto configuation information.
 */
static int prprobe(), prattach();

static struct mb_device *prinfo[1];		/* Supports only one */

/*
 * PRMAXVSIZE must be defined to be <= PRMAXSIZE.  This define
 * controls the maximum amount of virtual space that we are willing
 * to use for presto memory.  If too much space is used and there
 * are other devices present which take a lot of kernel virtual
 * address space (e.g. cgtwo), then bad things can happen.  Either
 * the system will perform rather poorly if the system is right at
 * the edge of its virtual space limitations or else it will hang
 * forever waiting for more virtual space to free up somehow.  If
 * we want to support larger presto memory sizes, then we might have
 * to consider not mapping everything in and just having a space of
 * virtual memory which we use as a ``mapping cache'' (ala seg_map).
 * You can use adb to patch prdriver+0x18 to adjust this limitation
 * from an object distribution.  After patching the /vmunix file,
 * just reboot the system and the new upper limit will be imposed.
 */

#define PRMAXVSIZE	0x100000

struct mb_driver prdriver = {
	prprobe, 0, prattach, 0, 0, 0,
	PRMAXVSIZE, "pr", prinfo, 0, 0, 0,
};

/*
 * The presto `ps' (presto status) auto configuation information.
 */
static int psprobe(), psattach();

static struct mb_device *psinfo[1];		/* Supports only one */
struct mb_driver psdriver = {
	psprobe, 0, psattach, 0, 0, 0,
	PSSIZE, "ps", psinfo, 0, 0, 0,
};

#ifdef COPY_PROTECT
/*
 * We store a magic number in the Dallas SMARTwatch so we can tell if the
 * nvram board was bought though us. We check both at probe time, and
 * some times in checkbatteries (in case someone attempted to get around
 * the copy protection by hacking the code to disable the check at probe time).
 *
 * The Dallas SMARTwatch is the creation of a hardware engineer who is
 * aggressively preserving address space and chip real estate.
 *
 * It hides behind a memory bank until a specific sequence of 64 bits is
 * written to bit 0 of its memory bank.
 *
 * Then bit 0 of the following 64 read/write cycles to that memory bank access
 * in a sequential fashion the bits of the SMARTwatch registers.
 *
 * Need I comment that this is a misguided hardware designed?
 */
#define CHECK_RATE	1000	/* how often to check smartwatch in checkbatteries */
#ifdef PRDEBUG
int pr_fakebadboard = 0;
#ifdef INIT_COPY_PROTECT
int set_watch = 1;	/* if 1, we will set the SMARTwatch to our magic num */
#else
int set_watch = 0;	/* if 1, we will set the SMARTwatch to our magic num */
#endif /* INIT_COPY_PROTECT */
#endif /* PRDEBUG */

u_int attention = 0x5ca33ac5;	/* the "magic" attention pattern */

struct smartwatch {	/* the SunOS compiler being "big indian" in layout */
		/* first word, high order bits first */
	int mode : 1;	/* 12 vs 24 hours mode, set = 12 hour mode 0-1 */
	int nc3  : 1;	/* not changeable */
	int dhour: 2;	/* decihours = tens of hours 0-2 */
	int hour : 4;	/* hours 0-9 */
	int nc2  : 1;	/* not changeable */
	int dmin : 3;	/* deciminutes = tens of minutes 0-5 */
	int min  : 4;	/* minutes 0-9 */
	int nc1  : 1;	/* not changeable */
	int dsec : 3;	/* deciseconds = tens of seconds 0-5 */
	int sec  : 4;	/* seconds 0-9 */
	int csec : 4;	/* centiseconds = .1 sec 0-9 */
	int msec : 4;	/* milliseconds = .01 sec 0-9 */
		/* second word, high order bits first */
	int year_h : 4;	/* high digit of year 0-9 */
	int year_l : 4;	/* low digit of year 0-9 */
	int nc6  : 3;	/* not changeable */
	int month_h : 1;	/* high digit of month 0-1 */
	int month_l : 4;	/* low digit of month 0-9 */
	int nc4  : 2;	/* not changeable */
	int day_of_month_h : 2;	/* high digit of day 0-3 */
	int day_of_month_l : 4;	/* low digit of day 0-9 */
	int nc5  : 2;	/* not changeable */
	int osc  : 1;	/* oscillator on/off, set = off 0-1 */
	int reset: 1;	/* reset ignore, set = ignore 0-1 */
	int day_of_week : 4;	/* 0-7 */
} smartwatch;

/*
 * Grab the attention of the Dallas SMARTwatch chip.
 *	This is done by sending a specific sequence of 64 bits to
 *	the memory bank the SMARTwatch is hiding behind.
 */
void
dsw_attention(dsw)
	register caddr_t dsw;
{
	register u_int bits;
	register int i;

	/*
	 * perform a read to reset the "bit pattern recognizer"
	 */
	bits = (unsigned int)*dsw;

	/*
	 * send the first set of 32 bits.
	 */
	bits = attention;
	for (i = 32; i > 0; i--, bits >>= 1) {
		*dsw = (unsigned char)bits;
#ifdef DSW_DEBUG
printf("%d", bits & 1);
#endif
	}
	/*
	 * send the second set of 32 bits.
	 */
	bits = attention;
	for (i = 32; i > 0; i--, bits >>= 1) {
		*dsw = (unsigned char)bits;
#ifdef DSW_DEBUG
printf("%d", bits & 1);
#endif
	}
#ifdef DSW_DEBUG
printf("\n");
#endif
}

/*
 * Write the SMARTwatch registers.
 *
 *	dsw is an address which the Dallas SMARTwatch is hiding behind.
 *	values is a pointer to a 64 bit object which will be used to fill
 *	the SMARTwatch registers.
 *
 *	To set the SMARTwatch registers, you trickle the bits down a bit at
 *	a time in bit 0.
 */
void
dsw_write(dsw, values)
	register caddr_t dsw;
	register unsigned int *values;
{
	register u_int bits;
	register int i;

	dsw_attention(dsw);	/* grab the attention of the SMARTwatch */

	bits = *values;
	for (i = 32; i > 0; i--, bits >>= 1) {
		*dsw = bits;
#ifdef DSW_DEBUG
printf("%d", bits & 1);
#endif
	}
	bits = *(values + 1);
	for (i = 32; i > 0; i--, bits >>= 1) {
		*dsw = bits;
#ifdef DSW_DEBUG
printf("%d", bits & 1);
#endif
	}
	/*
	 * Must perform a read to reset things again
	 * --Undocumented "feature" of the SMARTwatch.
	 */
	(void) peekc(dsw);
#ifdef DSW_DEBUG
printf("\n");
#endif
}

/*
 * Read the SMARTwatch registers.
 *	dsw is an address which the Dallas SMARTwatch is hiding behind.
 *	values is a pointer to a 64 bit object which will be filled with
 *	the SMARTwatch register contents
 *
 *	To read the SMARTwatch registers, you retrieve the bits a bit at
 *	a time from bit 0.
 */
void
dsw_read(dsw, values)
	register caddr_t dsw;
	register u_int values[];	/* 2 int's long */
{
	register u_int bits;
	register int i;

	dsw_attention(dsw);	/* grab the attention of the SMARTwatch */

	bits = 0;
	for (i = 0; i < 32; i++) {
		bits |= ((unsigned int)(*dsw & 0x1)) << i;
#ifdef DSW_DEBUG
printf("%d", (bits >> i) & 1);
#endif
	}
	values[0] = bits;
#ifdef PRDEBUG
	if (pr_fakebadboard)
		values[0] = 0x666;
#endif
	bits = 0;
	for (i = 0; i < 32; i++) {
		bits |= ((unsigned int)(*dsw & 0x1)) << i;
#ifdef DSW_DEBUG
printf("%d", (bits >> i) & 1);
#endif
	}
#ifdef DSW_DEBUG
printf("\n");
#endif
	values[1] = bits;
#ifdef PRDEBUG
	if (pr_fakebadboard)
		values[1] = 0x666;
#endif
}
#endif /* COPY_PROTECT */

#ifdef PRDEBUG
/*
 * presto buffer flags which should always be off.
 */
static int prclrflags = (B_READ|B_PHYS|B_ASYNC|B_FREE);

/*
 * presto buffer flags which should always be on.
 */
static int prsetflags = 0;

static void
checkbp(bp, str, bfset, bfclr)
	register struct buf *bp;
	char *str;
	int bfset, bfclr;
{

	if (((bp->b_flags & B_BUSY) == 0) && (bp->av_back->av_forw != bp ||
	    bp->av_forw->av_back != bp)) {
		printf("%s: free list bp=0x%x\n", str, bp);
		call_debug("freelist");
	}
	if (bp->b_back->b_forw != bp ||
	    bp->b_forw->b_back != bp) {
		printf("%s: hash list bp=0x%x\n", str, bp);
		call_debug("hashlist");
	}
	if ((bp->b_flags & (prclrflags | bfclr)) != 0) {
		printf("%s: clr flags = 0x%x, bp = 0x%x\n",
		    str, bp->b_flags, bp);
		call_debug("bad flags on");
	}
	if ((bp->b_flags & (prsetflags | bfset)) != (prsetflags | bfset)) {
		printf("%s: set flags = 0x%x, bp = 0x%x\n",
		    str, bp->b_flags, bp);
		call_debug("bad flags off");
	}
}

static void
prchk()
{
	register struct buf *bp, *dp;
	register struct bufhd *bhp;
	struct bufhd *ebhp;
	int fl, nbp;
	int allfree;
	int dirty, clean, inval;

	/*
	 * Verify all the buffers.
	 */
	allfree = dirty = clean = inval = 0;
	for (fl = 0; fl < PRBQUEUES; fl++) {
		dp = (struct buf *)&prbfreelist[fl];

		/*
		 * Verify the header bp;
		 */
		if (dp->b_flags != B_HEAD || dp->b_forw != dp ||
		    dp->b_back != dp) {
			printf("bad freelist header bp = 0x%x\n", dp);
			call_debug("bad header bp");
		}
		for (bp = dp->av_forw, nbp = 0; bp != dp;
		    bp = bp->av_forw, nbp++) {
			switch (fl) {
			case PR_DIRTY:
				checkbp(bp, "freelist",
				    B_DELWRI, B_INVAL | B_BUSY | B_ACTIVE);
				dirty++;
				break;
			case PR_CLEAN:
				checkbp(bp, "freelist",
				    0, B_INVAL | B_BUSY | B_ACTIVE);
				clean++;
				break;
			case PR_INVAL:
				checkbp(bp, "freelist",
				    B_INVAL, B_BUSY | B_ACTIVE);
				inval++;
				break;
			default:
				panic("prchk free");
				/* NOTREACHED */
			}
		}
		allfree += nbp;
	}

	PRDEBUG1("<free ok [%d %d %d] = %d> ",
	    prstatus.pr_ndirty, prstatus.pr_nclean,
	    prstatus.pr_ninval, allfree);

	if (dirty != prstatus.pr_ndirty || clean != prstatus.pr_nclean ||
	    inval != prstatus.pr_ninval) {
		printf("counted [%d %d %d]\n");
		call_debug("wrong buf counts");
	}

	ebhp = &prbufhash[PRBUFHSZ];
	for (bhp = prbufhash; bhp < ebhp; bhp++) {
		/*
		 * Verify the header bp;
		 */
		if (bhp->b_flags != B_HEAD) {
			printf("bad hashlist header bhp = 0x%x\n", bhp);
			call_debug("bad hash header bhp");
		}

		/*
		 * Walk each hash list
		 */
		dp = (struct buf *)bhp;
		for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
			checkbp(bp, "hashlist", 0, 0);
		}
	}
	PRDEBUG1("<hash ok> ", 0, 0, 0, 0);

	if (prcheck > 1) {
		/*
		 * Now walk each buffer...
		 */
		for (bp = prbufs; bp < &prbufs[prnbufs]; bp++) {
			checkbp(bp, "buf array", 0, 0);
		}
		PRDEBUG1("<bufs ok> ", 0, 0, 0, 0);
	}

	for (fl = 0; fl < prstatus.pr_nactive; fl++) {
		if (practive[fl] == NULL)
			call_debug("active list too short");
	}
	if (practive[fl] != NULL)
		call_debug("active list too long");
	PRDEBUG1("<%d active>\n", prstatus.pr_nactive, 0, 0, 0);
}
#endif /* PRDEBUG */

static prstates laststate;

/*
 * Goto error state, saving the previous
 * presto state in the laststate variable.
 * Will use prflushdev(NODEV, TRUE) to flush
 * back and inval dirty buffers, so no buffer
 * should be held by the caller or prdisable().
 */
static void
prdisable()
{

	if (prstatus.pr_state == PRERROR)
		return;
	laststate = prstatus.pr_state;
	prstatus.pr_state = PRERROR;
	if (laststate == PRUP)
		printf("presto: disabling...\n");
	prreadonly = 1;
	(void) prflushdev(NODEV, TRUE);
}

/*
 * Restore presto state to laststate
 * if we were in the error state.
 */
static void
prenable()
{

	if (prstatus.pr_state != PRERROR)
		return;

	if (laststate == PRUP) {
		printf("presto: back online!\n");
	}
	prreadonly = 0;
	prstatus.pr_state = laststate;
}

#if defined(SUN_RELEASE) && SUN_RELEASE != 0
#include <machine/cpu.h>

#define CHKSUM_CRC			/* use full CRC checksum */
#define PRCHKSUMDEFAULT		2	/* checksum absolutely everything */
#endif /* SUN_RELEASE */

/*
 * Software checksumming code to appease those who are worried about this
 * when all the other static memory in the system (page/segment/region
 * registers, data caches, and IO caches) are static memory w/o any parity.
 */

#define PRCHKSUMDISABLED	0	/* no check summing performed */
#define PRCHKSUMINOUT		1	/* verify on copyout (bounceio/hits) */
#define PRCHKSUMALL		2	/* verify on all outgoing buffers */
#define PRCHKSUMERROR		99	/* check sum error detected state */

#ifndef PRCHKSUMDEFAULT
#define PRCHKSUMDEFAULT		1	/* do checksumming on copyout */
#endif

static int prdochksum = PRCHKSUMDEFAULT;
static int prchksum_ignore;		/* true if rebooting w/ diff chksums */

#ifndef CHKSUM_CRC
#define CHKSUM_VERS		((prdochksum != 0) ? 1 : 0)
#else
#define CHKSUM_VERS		((prdochksum != 0) ? 2 : 0)
#endif


/*
 * Verify a buffers checksum without bcopying it anywhere.  This
 * routine can be used to verify buffer's content's w/o doing bounceio.
 * Returns TRUE if the buffer's checksum checks out correctly.
 *
 * Assumes that the addr pointers are int word aligned
 * values and that count is a multiple of PRFSIZE.
 */
static int
prverifybuf(busaddr, count)
	caddr_t busaddr;
	u_int count;
{
	register u_int *fromip;
	register u_int chksum, *chksump;
	register int lcnt, fcnt;

	if (prdochksum == PRCHKSUMDISABLED || prchksum_ignore)
		return (TRUE);			/* all ok */

	chksump = &prchksums[(busaddr - &nvbdp->nvbd_data[0][0])/PRFSIZE];

	ASSERT((count & (PRFSIZE - 1)) == 0);
	ASSERT(((int)busaddr & (sizeof (u_int) - 1)) == 0);

	for (lcnt = count / PRFSIZE, fromip = (u_int *)busaddr; lcnt > 0;
	    lcnt--, chksump++) {
#ifdef CHKSUM_CRC
#define UNROLLCNT	4			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += *fromip++;
		}
#else /* CHKSUM_CRC */
#define UNROLLCNT	16			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
			chksum += *fromip++;
		}
#endif /* CHKSUM_CRC */
#undef UNROLLCNT
		if (*chksump != chksum) {
			PRDEBUG1("presto: bad verify checksum (0x%x != 0x%x)\n",
			    *chksump, chksum, 0, 0);
			return (FALSE);
		}
	}
	return (TRUE);
}

/*
 * Error handling routine called when a presto checksum error is found.
 * We set the prdochksum global to an error state and then try and flush
 * and invalidate everything.  When we are in the checksum error state,
 * prbwrite() will validate each buffer by calling prverifybuf() before
 * doing the writeback.  Any corrupted buffers will be invalidated while
 * buffers that still look good will be synced back to the real disk.
 */
static void
prchksumerror(where)
	char *where;
{
	register struct buf *bp;
	struct buf *ebp;
	u_int nbadchksum, nwrite, nbusy;
	extern int panic_opt;

	printf("presto: checksum error detected during %s\n", where);

	prdochksum = PRCHKSUMERROR;
	prstatus.pr_state = PRERROR;
	prreadonly = 1;

	printf("scanning presto buffers for additional checksum errors...\n");

	/*
	 * Loop over all the presto buffers
	 */
	nbadchksum = nwrite = nbusy = 0;
	for (ebp = &prbufs[prnbufs], bp = prbufs; bp < ebp; bp++) {
		if ((bp->b_flags & (B_ACTIVE|B_INVAL|B_BUSY)) == B_INVAL)
			continue;		/* already on inval list */

		/*
		 * If the checksum looks bad,
		 * invalidate the entire buffer.
		 */
		if (!prverifybuf(bp->b_un.b_addr, bp->b_bcount)) {
			nbadchksum++;
			if ((bp->b_flags & B_BUSY) == 0) {
				/*
				 * Not busy, but now invalid and
				 * no longer has any useful data.
				 */
				prbremfree(bp);
				bp->b_flags = B_INVAL;
				PRDEBUG1(
			"invalidating corrupt non-busy buffer %x <%x, %x>\n",
				    bp, bp->b_un.b_addr, bp->b_bcount, 0);
				prbrelse(bp);
			} else {
				/*
				 * Currently "busy", but had a
				 * bad checksum.  Just sync back
				 * to NV RAM the fact that this
				 * bp is no longer valid.
				 */
				bp->b_flags = B_INVAL;
				PRDEBUG1(
			"invalidating corrupt busy buffer %x <%x, %x>\n",
				    bp, bp->b_un.b_addr, bp->b_bcount, 0);
				bptonvbp(bp);
			}
		} else {
			/* buffer looks fine */
			if ((bp->b_flags & B_BUSY) == 0) {
				prbremfree(bp);
				bp->b_flags |= B_INVAL;
				if ((bp->b_flags & B_DELWRI) == 0) {
					prbrelse(bp);
				} else {
					nwrite++;
					prbwrite(bp);
				}
			} else {
				nbusy++;
				PRDEBUG1(
			"skipping busy valid buffer %x <%x, %x>\n",
				    bp, bp->b_un.b_addr, bp->b_bcount, 0);
			}
		}
	}

	printf("presto: invalidated %d buffer%s with bad checksums\n",
	    nbadchksum, nbadchksum == 1 ? "" : "s");

	while (prstatus.pr_nactive > 0) {
		/*
		 * Wait on the next buffer until active list is drained.
		 * Normally this list already empty as prbwrite() works
		 * synchronously when presto is in the error state.
		 */
		prbwait(*practive);
	}

	printf("presto: flushed %d valid buffer%s (%d busy)\n",
	    nwrite, nwrite == 1 ? "" : "s", nbusy);

	/*
	 * Now panic.  We set a global flag to avoid trying to
	 * sync the filesystem as this process probably holds a
	 * locked inode (since it was probably doing a synchronous
	 * write from bmap() or rwip()).  On systems like SunOS 4.x
	 * where a fully logically cache is maintained, this locked
	 * inode will probably be needed to perform logical to physical
	 * translations needed to perform the sync operation and
	 * we don't want to hang waiting on reboot.  On systems
	 * that still use a physical cache for local file system,
	 * this hack doesn't need to be used.
	 */
	panic_opt |= RB_NOSYNC;
	panic("presto checksum error");
	/* NOTREACHED */
}

/*
 * Fill up the given presto buffer, calculating and
 * storing the checksum for each PRFSIZE chunk in the
 * prchksums[] array if checksumming is enabled.
 *
 * Assumes that the addr pointers are int word aligned
 * values and that count is a multiple of PRFSIZE.
 */
static void
prnvbufwrite(busaddr, vaddr, count)
	caddr_t busaddr;
	caddr_t vaddr;
	u_int count;
{
	register u_int *toip, *fromip;
	register u_int chksum, *chksump;
	register int lcnt, fcnt;

	if (prdochksum == PRCHKSUMDISABLED) {
		BUSWRITE(busaddr, vaddr, count);
		return;
	}

	chksump = &prchksums[(busaddr - &nvbdp->nvbd_data[0][0])/PRFSIZE];

	ASSERT((count & (PRFSIZE - 1)) == 0);
	ASSERT(((int)busaddr & (sizeof (u_int) - 1)) == 0);
	ASSERT(((int)vaddr & (sizeof (u_int) - 1)) == 0);

	for (lcnt = count / PRFSIZE, toip = (u_int *)busaddr,
	    fromip = (u_int *)vaddr; lcnt > 0; lcnt--) {
		register u_int t;

#ifdef CHKSUM_CRC
#define UNROLLCNT	4			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
		}
#else /* CHKSUM_CRC */
#define UNROLLCNT	16			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
		}
#endif /* CHKSUM_CRC */
#undef UNROLLCNT
		*chksump++ = chksum;
	}
}

/*
 * Copy out from the given presto buffer, calculating and
 * verifying the checksum for each PRFSIZE chunk in the
 * prchksums[] array if checksumming is enabled.  This
 * gives us the effect of parity checking in software with
 * minimal performance impact since we do this when filling
 * and draining the buffers themselves.
 *
 * Assumes that the addr pointers are int word aligned
 * values and that count is a multiple of PRFSIZE.
 */
static void
prnvbufread(busaddr, vaddr, count)
	caddr_t busaddr;
	caddr_t vaddr;
	u_int count;
{
	register u_int *toip, *fromip;
	register u_int chksum, *chksump;
	register int lcnt, fcnt;

	if (prdochksum == PRCHKSUMDISABLED || prchksum_ignore) {
		BUSREAD(busaddr, vaddr, count);
		return;
	}

	chksump = &prchksums[(busaddr - &nvbdp->nvbd_data[0][0])/PRFSIZE];

	ASSERT((count & (PRFSIZE - 1)) == 0);
	ASSERT(((int)busaddr & (sizeof (u_int) - 1)) == 0);
	ASSERT(((int)vaddr & (sizeof (u_int) - 1)) == 0);

	for (lcnt = count / PRFSIZE, toip = (u_int *)vaddr,
	    fromip = (u_int *)busaddr; lcnt > 0; lcnt--, chksump++) {
		register u_int t;

#ifdef CHKSUM_CRC
#define UNROLLCNT	4			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
			t = *fromip++;
			if (chksum & 1)
				chksum = (chksum >> 1) + 0x80000000;
			else
				chksum >>= 1;
			chksum += t;
			*toip++ = t;
		}
#else /* CHKSUM_CRC */
#define UNROLLCNT	16			/* count of unrolled moves */
		for (chksum = 0, fcnt = PRFSIZE / sizeof (u_int); fcnt > 0;
		    fcnt -= UNROLLCNT) {
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
			t = *fromip++; chksum += t; *toip++ = t;
		}
#endif /* CHKSUM_CRC */
#undef UNROLLCNT
		if (*chksump != chksum) {
			PRDEBUG1("presto: bad checksum (0x%x != 0x%x)\n",
			    *chksump, chksum, 0, 0);
			prchksumerror("read back");
		}
	}
}

/*
 * Clear all the presto IO statistics
 */
static void
clearstats()
{

	bzero((caddr_t)&prstatus.pr_rdstats, sizeof (prstatus.pr_rdstats));
	bzero((caddr_t)&prstatus.pr_wrstats, sizeof (prstatus.pr_wrstats));
	pr_stats_cleared = (u_long)time.tv_sec;
}

/*
 * Sync back the given presto buffer to the non-volatile memory nvbuf.
 * If the given bp is zero, sync back all the presto buffers.
 */
static void
bptonvbp(bp)
	register struct buf *bp;
{
	register struct nvbuf *nvbp;
	register int i, nbp;
	struct nvbuf nvb;

	if (bp != NULL) {
		nvbp = &nvhp->nv_bufs[bp - prbufs];
		nbp = 1;
	} else {
		bp = prbufs;
		nvbp = nvhp->nv_bufs;
		nbp = prnbufs;
	}

	for (i = 0; i < nbp; i++) {
		nvb.nb_flags = bp->b_flags;
		nvb.nb_blkno = bp->b_blkno;
		nvb.nb_dev = bp->b_dev;
		nvb.nb_bcount = bp->b_bcount;
		BUSWRITE((caddr_t)nvbp, (caddr_t)&nvb, sizeof (nvb));
		bp++;
		nvbp++;
	}

	/*
	 * If we have an "dirty LED" to update,
	 * load up the current dirty status.
	 */
	if (prdirty_led != NULL) {
		*prdirty_led = (prstatus.pr_ndirty != 0 ||
				prstatus.pr_nactive != 0);
	}
}

/*
 * Handle presto size dependent actions for onboard memory.
 * This includes setting the prdesavail and prnbufs variables
 * and reallocating any dynamic data structures (e.g., prbufs
 * and practive) which depend on the size of the presto memory.
 * Then the presto buffers, free lists, hash lists, and the presto
 * status data structure are initialized for all invalid buffers.
 */
static void
prinit(size)
	u_int size;
{
	register struct diskhd *fl;
	register struct buf *bp;
	struct buf *ebp;
	register struct bufhd *bhp;
	struct bufhd *ebhp;
	register caddr_t mp;

	if (size < PRMINSIZE || size > prstatus.pr_maxsize ||
	    prstatus.pr_state == PRUP) {
		panic("prinit");
		/* NOTREACHED */
	}

	if (prnbufs != 0) {
		/*
		 * Free up last allocations.
		 */
		kmem_free((caddr_t)prbufs, (sizeof (*prbufs) * prnbufs));
#ifndef USE_B_CHAIN
		kmem_free((caddr_t)proldflags,
		    (sizeof (*proldflags) * prnbufs));
#endif
		kmem_free((caddr_t)practive,
		    (sizeof (*practive) * (prnbufs + 1)));
	}

	/*
	 * Perform sizing operations.
	 */
	prnbufs = (size - ((caddr_t)nvbdp - (caddr_t)praddr)) / PRBSIZE;
	prdesavail = MIN(prdesnavail, MAX(prnbufs / 5, 1));

	/*
	 * Perform new allocations.
	 */
	prbufs = (struct buf *)kmem_zalloc((sizeof (*prbufs) * prnbufs));
#ifndef USE_B_CHAIN
	proldflags = (int *)kmem_zalloc((sizeof (*proldflags) * prnbufs));
#endif
	practive = (struct buf **)
	    kmem_zalloc((sizeof (*practive) * (prnbufs + 1)));

	/*
	 * Initial the data structures.
	 */
	prstatus.pr_cursize = size;
	prstatus.pr_ninval = prnbufs;
	prstatus.pr_ndirty = prstatus.pr_nclean = prstatus.pr_nactive = 0;

	/*
	 * Initialize free lists
	 */
	for (fl = prbfreelist; fl < &prbfreelist[PRBQUEUES]; fl++) {
		fl->b_forw = fl->b_back = fl->av_forw = fl->av_back =
		    (struct buf *)fl;
		fl->b_flags = B_HEAD;
	}

	/*
	 * Initialize hash chains
	 */
	ebhp = &prbufhash[PRBUFHSZ];
	for (bhp = prbufhash; bhp < ebhp; bhp++) {
		bhp->b_forw = bhp->b_back = (struct buf *)bhp;
		bhp->b_flags = B_HEAD;
	}

	/*
	 * Initialize buffers
	 */
	fl = &prbfreelist[PR_INVAL];
	mp = &nvbdp->nvbd_data[0][0];
	ebp = &prbufs[prnbufs];
	for (bp = prbufs; bp < ebp; bp++) {
		bp->b_flags = B_INVAL;
		bp->b_bcount = bp->b_bufsize = PRBSIZE;
		bp->av_forw = bp->av_back = bp->b_forw = bp->b_back = bp;
		binsheadfree(bp, (struct buf *)fl);
		bp->b_un.b_addr = mp;
		mp += PRBSIZE;
	}
	clearstats();
}

/*
 * Initialize the non-volatile presto memory data structures.
 */
static void
prnvinit(size)
	u_int size;
{
	struct nvh nvh;

	nvh.nvh_size = size;
	nvh.nvh_magic = PRMAGIC;
	nvh.nvh_version = PRVERSION;
	nvh.nvh_machineid = machineid();
	nvh.nvh_bsize = PRBSIZE;
	nvh.nvh_fsize = PRFSIZE;
	nvh.nvh_nbufs = prnbufs;
	nvh.nvh_chksum = CHKSUM_VERS;

	BUSZERO((caddr_t)nvhp, size - ((u_int)nvhp - (u_int)praddr));
	BUSWRITE((caddr_t)&nvhp->nv_nvh, (caddr_t)&nvh, sizeof (struct nvh));

	/*
	 * Sync back all the buffer contents of
	 * on board memory with presto memory.
	 */
	bptonvbp((struct buf *)0);

	PRCHECK();
}

/*
 * Handle putting the given presto buffer back on the right free list.
 * Note that the buffer remains on any hash list it was on on except
 * when the buffer is marked with B_INVAL.
 */
static void
prbrelse(bp)
	register struct buf *bp;
{
	register struct diskhd *flist;
#ifdef USE_B_CHAIN
	int oldflags;
#endif

	PRCHECK();

	if ((bp->b_flags & B_WANTED) != 0)
		wakeup((caddr_t)bp);

	if ((bp->b_flags & B_ERROR) != 0) {
		/*
		 * We had a write error, mark the buffer as dirty again.
		 */
		bp->b_flags |= B_DELWRI;
		bp->b_flags &= ~B_INVAL;
	}

	if ((bp->b_flags & B_DELWRI) != 0) {
		flist = &prbfreelist[PR_DIRTY];
		prstatus.pr_ndirty++;
	} else if ((bp->b_flags & B_INVAL) != 0) {
		bremhash(bp);
		bp->b_forw = bp->b_back = bp;
		bp->b_dev = 0;
		flist = &prbfreelist[PR_INVAL];
		prstatus.pr_ninval++;
	} else {
		flist = &prbfreelist[PR_CLEAN];
		prstatus.pr_nclean++;
	}
	binstailfree(bp, (struct buf *)flist);
	bp->b_flags &= ~(B_WANTED | B_BUSY | B_ACTIVE | B_DONE);

	/*
	 * Commit new buffer state to nv memory if something has changed.
	 */
#ifdef USE_B_CHAIN
	oldflags = (int)bp->b_chain;
	if (oldflags != bp->b_flags) {
		bptonvbp(bp);
	}
#else
	if (proldflags[bp - prbufs] != bp->b_flags) {
		bptonvbp(bp);
	}
#endif
	PRCHECK();
}

/*
 * Start a write back operation on a bp,
 * queueing it up on the practive list.
 *
 * N.B. - on architectures that need special help to allow a random
 * disk controller to address S/VMEbus memory something will have to be
 * added here.  On Sun's, the mb routines will notice that S/VMEbus
 * memory is mapped at the kernel virtual address for the IO and
 * will make sure that the controller will directly S/VMEbus space
 * instead of using DVMA.  Unfortunately, some SCSI host adapter
 * software stupidly assume we must be using DVMA and in some cases
 * try to touch DVMA space directly to compensate for stupid hardware
 * that doesn't always leave things in a nice state.  The SCSI `si', `se',
 * and `sw' drivers in pre 4.1 SunOS systems are guilty of this problem.
 * The `sc' driver does this only for odd byte transfers in pre 4.1 SunOS
 * systems, which will we will never generate so we are ok here.
 */
static void
prbwrite(bp)
	register struct buf *bp;
{
	register struct prtab *prtab;
	dev_t maj_d, min_d;

	maj_d = major(bp->b_dev);
	min_d = minor(bp->b_dev);

	ASSERT(PR_MAJORBDEV_OK(maj_d));
	ASSERT(isclr(prtabs[maj_d]->pt_error.bits, min_d));

	if ((bp->b_flags & (B_ACTIVE|B_DELWRI)) != B_DELWRI) {
		panic("prbwrite");
		/* NOTREACHED */
	}

	PRTRACE("<write dev %x, blkno %x> ", bp->b_dev, bp->b_blkno);
	bp->b_flags |= (B_BUSY | B_ACTIVE);
	bp->b_flags &= ~(B_DONE | B_DELWRI | B_ERROR);
	/*
	 * We don't have to sync new flag to NV memory - an
	 * "active buffer" is the same as a "dirty buffer" upon reboot.
	 */

	/*
	 * See if we have too many outstanding writes right now.
	 */
	while (prstatus.pr_nactive >= pr_write_limit) {
		(void) prreap();
		if (prstatus.pr_nactive >= pr_write_limit)
			(void) sleep((caddr_t)&pr_write_limit, PRIBIO);
	}
	practive[prstatus.pr_nactive++] = bp;		/* add to active list */

	prtab = prtabs[maj_d];
	if (BOUNCEIO(prtab, min_d)) {
		register caddr_t buf;

		/*
		 * Use a buffer from non vme space to hand the data
		 * to a device which cannot access vme space for io.
		 */
		while (pr_bounce_active >= pr_bounce_limit) {
			(void) prreap();
			if (pr_bounce_active >= pr_bounce_limit)
				(void) sleep((caddr_t)&pr_bounce_limit, PRIBIO);
		}
		buf = kmem_fast_alloc(&prfastpool, PRBSIZE, 1);
		prnvbufread(bp->b_un.b_addr, buf, bp->b_bcount);
		bp->b_un.b_addr = buf;
		pr_bounce_active++;
		if (pr_bounce_active > pr_bounce_max)
			pr_bounce_max = pr_bounce_active;
#ifndef SUNSCRUB
	} else if (prdochksum == PRCHKSUMALL &&
	    !prverifybuf(bp->b_un.b_addr, bp->b_bcount)) {
		/*
		 * A checksum error was found - call the chksum error
		 * handling routine after putting the buffer back
		 * on the dirty freelist by marking it with an error.
		 */
		practive[--(prstatus.pr_nactive)] = NULL;
		bp->b_flags |= B_ERROR;
		prbrelse(bp);
		prchksumerror("write back");
		/* NOTREACHED */
	}
#else
	/*
	 * BSA: Sun guaranteed scrubbing of buffers, See comment at top
	 * After every n (n=50, above) writes, it's time to do a checksum
	 */
	} else if (prdochksum == PRCHKSUMALL) {
#ifdef SUNSCRUB_BIG_DEBUG
printf("Presto: incrementing I/O count, %d\n", pr_ios);
#endif
		if ((++pr_ios >= pr_nth) && 		/* Time for a scrub ? */
		    (pr_nth > 0) &&			/* To turn this off */
		    ((prstatus.pr_ndirty > 0) ||	/* Any bufs to scrub? */
		     (prstatus.pr_nclean > 0))) {
			/*
			 * Skip bufs that have B_INVAL flag set, not in use.
			 * NOTE: If we hit a B_ACTIVE buffer, we should still
			 * be able to do a checksum on it, as even if it's been
			 * passed to a physical device to write-thru, 
			 * that device will be accessing it as read only.
			 * BUT, I'm going to be paranoid & skip B_ACTIVE bufs
			 */
			struct buf *scrub;
			struct buf *start;

#ifdef SUNSCRUB_BIG_DEBUG
printf("presto: Start next scrub @ buffer %d ...\n", pr_next_scrub);
#endif
			scrub = prbufs + pr_next_scrub;
			start = scrub;
			while (((scrub->b_flags & (B_INVAL | B_ACTIVE)) != 0)) {
				if (++pr_next_scrub >= prnbufs)
					pr_next_scrub = 0;
				/*
				 * OK, I'm paranoid. Check of prstatus above
				 * should guarantee, we'll never loop around
				 */
				if (((scrub=prbufs+pr_next_scrub)) == start)
					break;
			}
#ifdef SUNSCRUB_DEBUG
printf("presto: Scrub buffer %d", pr_next_scrub);
#ifdef SUNSCRUB_BIG_DEBUG
printf(", (flags=%x)", scrub->b_flags);
#endif
printf("\n");
if (bp == scrub)
	printf("presto: scrub buffer same as current write buffer\n");
#endif
			if (((scrub->b_flags & (B_INVAL | B_ACTIVE)) != 0)) { 
#ifdef SUNSCRUB_DEBUG
printf("presto: error scrub on (B_INVAL | B_ACTIVE) buffer, skipping\n");
#endif
			}
			else if (!prverifybuf(scrub->b_un.b_addr, 
						scrub->b_bcount)) {
				/*
				 * BSA: If possible, should probably just warn
				 * that this buffer is bad & mark it unuseable
				 * instead of panic'ing below. Since we should
				 * never see checksum error, not going to worry
				 */
#ifdef SUNSCRUB_DEBUG
printf("presto: checksum error detected\n");
#else
				prchksumerror("scrub buffer");
				/* NOTREACHED */
#endif
			}	/* End of checksum bad for scrubbed buffer */

			/* Increment/reset our counters */
			pr_ios = 0;
			if (++pr_next_scrub >= prnbufs)
				pr_next_scrub = 0;
			
		}	/* End of time to do next scrub check */
	}	/* End of writeback checksum/buffer scrub section */
#endif

#ifdef PRDEBUG
	prblkno[prblkno_i++ & 0xff] = bp->b_blkno;
#endif /* PRDEBUG */
	(*prtab->pt_strategy)(bp);

	/*
	 * If we are in the error state, wait for the write to complete,
	 * and then reap it so we can notice when io can again succeed.
	 * In the same vain, if we are up against some write limit
	 * (either bounceio or total number of outstanding writes),
	 * then we wait for the write to complete here (degrading to
	 * synchronous mode operation).  This way we can be sure to
	 * the do any needed wakeups from this process immediately
	 * after the IO is done without having to depend on another
	 * call to prbwrite() being made to wake the blocked processes.
	 */
	if (prstatus.pr_state == PRERROR ||
	    prstatus.pr_nactive >= pr_write_limit ||
	    (BOUNCEIO(prtab, min_d) && pr_bounce_active >= pr_bounce_limit)) {
		prbwait(bp);
		(void) prreap();
	}
}

/*
 * Set up the battery mode and variables.  Also reset batterystatus value.
 */
void
set_battery_mode(mode)
	enum battery_modes mode;
{
	int i;

	if (fakebattery != 0) {
		battery_mode = BAT_FAKE;
	} else {
		battery_mode = mode;
	}

	/*
	 * For MM1880 style boards, the battery
	 * count comes from magic number.
	 */
	if (battery_mode != BAT_MM1800)
		prstatus.pr_battcnt = PS_BATTCNT;

	/*
	 * Set up variables based on the number of actual batteries.
	 */
	Pr_allbattenabled = Pr_allbattlow = 0;
	for (i = 0; i < prstatus.pr_battcnt; i++) {
		if (battery_mode != BAT_MM6704)
			Pr_allbattenabled |= PR_BATTENABLED(i);
		Pr_allbattlow |= PR_BATTLOW(i);
	}
	Pr_dontcare = ~(Pr_allbattenabled | Pr_allbattlow);
	batterystatus = Pr_normal = Pr_allbattenabled;
}

/*
 * Check the current battery status and if different from the last
 * read status, print out new information and initialize the global
 * batterystatus and status.pr_battlow[] variables appropriately.
 * If all of the batteries are low, then we shutdown presto to
 * avoid losing data if we lose power to the system.  This routine
 * should not be called with any presto buffers held since we will
 * call prdisable() if the battery status changes to all low.
 *
 * Returns non-zero if all batteries are low or all disabled.
 */
static int
checkbatteries()
{
	register u_char bstatus;
	u_char oldbatterystatus;
	register int i;
	static int battery_check = 0;
	static int battery_bound;

	/*
	 * How are we to check the batteries ?
	 */
	switch (battery_mode) {
		case BAT_NONE:
			batterystatus = Pr_allbattlow;
			return (1);

		case BAT_BAD:
			bstatus = Pr_allbattlow;
			break;

		case BAT_FAKE:
			bstatus = fakebattery;
			break;

		case BAT_PRETEND:
			/*
			 * No presto status device to read,
			 * just pretend all is ok.
			 */
			bstatus = Pr_normal;
			break;

		case BAT_MM6704:
#ifdef COPY_PROTECT
			if (battery_check++ >= battery_bound) {
				if (battery_bound == 0) {
					battery_bound = CHECK_RATE +
						(time.tv_usec & 0x2FF);
				} else {
					dsw_read((caddr_t)&psaddr->mem_clock,
						(u_int *)&smartwatch);
#ifdef DSW_DEBUG
					/*
					 * So what was it set to?
					 */
					printf(
				"day %d:%d, month %d:%d, year %d:%d, osc %d\n",
						smartwatch.day_of_month_h,
						smartwatch.day_of_month_l,
						smartwatch.month_h,
						smartwatch.month_l,
						smartwatch.year_h,
						smartwatch.year_l,
						smartwatch.osc);
#endif /* DSW_DEBUG */
					/*
					 * Verify the magic number stored
					 * in the SMARTwatch clock.
					 */
					if (smartwatch.day_of_month_h != 0 ||
					    smartwatch.day_of_month_l != 7 ||
					    smartwatch.month_h != 0 ||
					    smartwatch.month_l != 5 ||
					    smartwatch.year_h != 6 ||
					    smartwatch.year_l != 2) {
						printf("presto: bad battery status!\n");
						battery_mode = BAT_BAD;
						prdisable();
						return (batterystatus ==
							 Pr_allbattlow);
					}
					battery_check = 0;
					battery_bound = CHECK_RATE +
						(time.tv_usec & 0x2FF);
				}
			}
#endif /* COPY_PROTECT */
			/* fall though */
#ifdef sun
			bstatus = *battery_registers;
#else /* sun */
			{
				u_char tmp;

				BUSREAD(battery_registers, &tmp, sizeof (tmp));
				bstatus = tmp;
			}
#endif /* sun */
			bstatus &= ~Pr_dontcare;
			break;

		case BAT_MM1800:
#ifdef sun
			/* skip using macros for speed here... */
			bstatus = (u_char)*((u_long *)battery_registers);
#else /* sun */
			{
				u_long tmp;

				BUSREAD((u_long  *)battery_registers, &tmp,
				    sizeof (tmp));
				bstatus = (u_char)(tmp & 0xff);
			}
#endif /* sun */
			bstatus &= ~Pr_dontcare;
	}

	/*
	 * Initialize global variables.
	 */
	oldbatterystatus = batterystatus;
	batterystatus = bstatus;

	/*
	 * Only bother with updating prstatus and printf's if something changed.
	 */
	if ((bstatus ^ oldbatterystatus) != 0) {
		register int disable = 0;

		for (i = 0; i < prstatus.pr_battcnt; i++) {
			if ((bstatus & PR_BATTLOW(i)) != 0) {
				prstatus.pr_batt[i] = BAT_LOW;
			} else if (battery_mode != BAT_MM6704 &&
			    ((bstatus & PR_BATTENABLED(i)) == 0)) {
				prstatus.pr_batt[i] = BAT_DISABLED;
			} else {
				prstatus.pr_batt[i] = BAT_GOOD;
			}
		}
		printf("presto: ");
		for (i = 0; i < prstatus.pr_battcnt; i++) {
			printf("%sbattery #%d %s",
			    i == 0 ? "" : ", ", i + 1,
			    (bstatus & PR_BATTLOW(i)) ? "LOW" : "ok");
		}
		printf("\n");
		if ((bstatus & Pr_allbattlow) == Pr_allbattlow) {
			printf("presto: all batteries are low!\n");
			disable = 1;
		}
		if (battery_mode != BAT_MM6704) {
			printf("presto: ");
			for (i = 0; i < prstatus.pr_battcnt; i++) {
				printf("%sbattery #%d %s",
				    i == 0 ? "" : ", ", i + 1,
				    (bstatus & PR_BATTENABLED(i)) ?
					"enabled" : "DISABLED");
			}
			printf("\n");
			if ((bstatus & Pr_allbattenabled) == 0) {
				printf("presto: all batteries are disabled!\n");
				disable = 1;
			}
		}
		if (disable) {
			prdisable();
		} else if (prstatus.pr_state == PRERROR &&
		    prstatus.pr_ndirty == 0) {
			prenable();
		}
	}

	/*
	 * Returns non-zero if all batteries are low or none are enabled.
	 */
	return ((bstatus & Pr_allbattlow) == Pr_allbattlow ||
	    (battery_mode != BAT_MM6704 && (bstatus & Pr_allbattenabled) == 0));
}

/*
 * Reap all the active presto buffers that have been marked as done.
 * Returns true if we *might* have slept somewhere.
 */
static int
prreap()
{
	register struct buf **bpp, *bp;
	static int level = 0;
	dev_t maj_d, min_d;
	int err, maybe_slept;

	if (prstatus.pr_nactive == 0)
		return (0);

	/*
	 * We don't bother to spl while checking the B_DONE flag
	 * even thou it might be changed at interrupt level.
	 * The worse that happens is that we don't reap everything
	 * we might have been able to, but this isn't important here.
	 */
	maybe_slept = 0;
	for (bpp = practive; (bp = *bpp) != NULL; ) {
		if ((bp->b_flags & B_DONE) == 0) {
			bpp++;
			continue;
		}

		/*
		 * If at write limit, let someone else in.
		 */
		if (prstatus.pr_nactive >= pr_write_limit)
			wakeup((caddr_t)&pr_write_limit);

		*bpp = practive[--(prstatus.pr_nactive)];
		practive[prstatus.pr_nactive] = NULL;

		maj_d = major(bp->b_dev);
		min_d = minor(bp->b_dev);
		err = bp->b_flags & B_ERROR;
		if (err) {
			PRDEBUG1(
			    "prreap: error on dev (%d, %d) for blk %d\n",
			    maj_d, min_d, bp->b_blkno, 0);
		}

		/*
		 * Give the buffer back to the pool of staging buffers if
		 * we are not doing the io directly from the nvram in
		 * vme space.
		 */
		if (BOUNCEIO(prtabs[maj_d], min_d)) {
			/* verify that it's really a presto buffer header */
			ASSERT(bp - prbufs >= 0 &&
			    bp - prbufs < nvhp->nv_nvh.nvh_nbufs);
			/* verify that it's buffer is not in presto memory */
			ASSERT(bp->b_un.b_addr < praddr ||
			    bp->b_un.b_addr > praddr + nvhp->nv_nvh.nvh_size);
			kmem_fast_free(&prfastpool, bp->b_un.b_addr);
			bp->b_un.b_addr = nvbdp->nvbd_data[bp - prbufs];

			/* if at limit, somebody may be waiting for a buffer */
			if (pr_bounce_active >= pr_bounce_limit)
				wakeup((caddr_t)&pr_bounce_limit);
			pr_bounce_active--;
		}

		prbrelse(bp);

		/*
		 * See if we are changing state.  This happens when
		 * we are up and we get a disk error or when we are
		 * down on a given dev and we notice that something
		 * got through ok.
		 *
		 * We use the "level" variable to measure our
		 * recursion.  We need to be able to recursive more
		 * once to verify that a device is really up or down
		 * here.  But we cannot recursive forever with
		 * prflushdev(NODEV, TRUE) calling prbwait() calling prreap()..
		 */

		level++;
		if (err && prstatus.pr_state != PRERROR) {
			printf("presto: error on dev (%d, %d)\n",
			    maj_d, min_d);
			if (level == 1)
				prdisable();
		} else if (prstatus.pr_state == PRERROR && !err &&
		    prdochksum != PRCHKSUMERROR) {
			/*
			 * We don't bother with all of this recovery stuff
			 * if we have found a checksum error and we are
			 * currently trying to write back any valid buffers.
			 */
			PRDEBUG1("prreap[%d]: dev (%d, %d) write ok\n",
			    level, maj_d, min_d, 0);
			if (prstatus.pr_nactive == 0 &&
			    prstatus.pr_ndirty == 0) {
				if (prresetting) {
					printf(
					    "presto: dirty buffers written\n");
					prreset_finish();
				}
				if (checkbatteries() == 0) {
					/*
					 * We finally got everything out ok and
					 * the batteries look ok, we can come
					 * back up.
					 */
					prenable();
				}
			} else {
				PRDEBUG1("prreap[%d]: %d dirty bufs left\n",
				    level, prstatus.pr_ndirty, 0, 0);
				if (level == 1 && !prresetting &&
				    prflushdev(NODEV, TRUE)) {
					maybe_slept = 1;
				}
			}
		}
		--level;
	}
	return (maybe_slept);
}

/*
 * Wait for the given presto bp to be returned to the freelist.
 */
static void
prbwait(bp)
	register struct buf *bp;
{
	int s;

loop:
	if ((bp->b_flags & B_BUSY) == 0) {
		/*
		 * Already on free list, nothing to do.
		 */
		return;
	}

	if ((bp->b_flags & B_ACTIVE) == 0) {
		/*
		 * Someone just has the buffer checked out,
		 * but it hasn't been passed off to a driver.
		 */
		while ((bp->b_flags & B_BUSY) != 0) {
			bp->b_flags |= B_WANTED;
			PRDEBUG1("prbwait: busy wait bp %x, flags %x\n",
			    bp, bp->b_flags, 0, 0);
			(void) sleep((caddr_t)bp, PRIBIO);
		}
		return;
	}

	/*
	 * Buffer is busy for real IO.
	 */
	s = splbio();
	/* while busy for IO */
	while ((bp->b_flags & (B_ACTIVE | B_DONE)) == B_ACTIVE) {
		bp->b_flags |= B_WANTED;
		PRDEBUG3("io wait: buf=0x%x, flags=0x%x\n",
		    bp, bp->b_flags, 0, 0);
		(void) sleep((caddr_t)bp, PRIBIO);
	}
	(void) splx(s);

	/* if not reapable, start all over */
	if ((bp->b_flags & (B_ACTIVE | B_DONE)) != (B_ACTIVE | B_DONE)) {
		PRDEBUG1("some else reaped %x\n", bp, 0, 0, 0);
		goto loop;
	}

	/*
	 * IO is now done, use prreap() to reap the buffer.
	 */
	(void) prreap();
	if ((bp->b_flags & B_BUSY) != 0) {
		panic("prbwait");
		/* NOTREACHED */
	}
}

/*
 * Remove the given presto buffer from a free list.
 */
static void
prbremfree(bp)
	register struct buf *bp;
{

	bremfree(bp);
	PRFIXFREE(bp);
	/*
	 * Save away old value of flags.
	 * When this buffer is later free'd we will sync back the buffer
	 * to non-volatile memory only if the flags have been changed
	 * (e.g., the B_DELWRI or B_INVAL flags went on).
	 */
#ifdef USE_B_CHAIN
	bp->b_chain = (struct buf *)bp->b_flags;
#else
#ifdef PRDEBUG
	if ((prbufs - bp) >= prnbufs)
		printf("pr0: non presto buffer prbremfree'd. delta=0x%x\n",
			prbufs - bp);
	else
#endif
		proldflags[bp - prbufs] = bp->b_flags;
#endif

	bp->b_flags |= B_BUSY;
	if ((bp->b_flags & B_DELWRI) != 0) {
		prstatus.pr_ndirty--;
	} else if ((bp->b_flags & B_INVAL) != 0) {
		prstatus.pr_ninval--;
	} else {
		prstatus.pr_nclean--;
	}
}

/*
 * Called to synchronously flush back (and invalidate is inval is TRUE)
 * all buffers for the given dev.  If dev == NODEV, then all buffers are
 * flushed (and possibly invalidated).
 *
 * No flushing will occur on a device whose "open failed" bit is set.
 *
 * Return 1 if we might have slept, 0 otherwise.
 *
 * On return from this routine, all buffers will be flushed
 * (and invalidated if inval is TRUE) for the specified device(s),
 * or presto will be in the PRERROR state (and could have an
 * error bit set [failed open] for the device(s)).
 */
static int
prflushdev(dev, inval)
	dev_t dev;
	int inval;
{
	register struct buf *bp, *ebp;
	register struct prtab *prtab;
	register int maybe_slept = 0;
	register int scan_needed;
	dev_t maj_d, min_d;
	int passno;

	if (prbufs == NULL) {
		return (0);		/* no presto buffers to flush */
	}

	maj_d = major(dev);
	min_d = minor(dev);

	/*
	 * Loop until we have made a complete pass without
	 * sleeping or returning in an error condition.
	 * If we slept during a pass, another process can enter
	 * the this and released our write lock on this pass.
	 * So we loop until we can get a hopefully non-sleeping
	 * scan to verify that everything is written.
	 */
	for (passno = 0, scan_needed = 1; scan_needed; passno++) {
		scan_needed = 0;

		if (dev == NODEV) {
			/*
			 * Are global flushes currently allowed
			 * (i.e. are the error bits valid)?
			 * If not, then just get out now.
			 */
			if (prresetting || prresetneeded) {
				/* ASSERT(prstatus.pr_state == PRERROR); */
				return (0);
			}
			prreadonly = 1;
		} else {
			prtab = prtabs[maj_d];

			/*
			 * Verify IO is allowed on this device.
			 * This isn't strictly needed if we are only
			 * called after a success internal open in
			 * prreset() or if we have are called with a
			 * device that was successfully opened else where.
			 */
			if (isset(prtab->pt_error.bits, min_d))
				return (0);
			setbit(prtab->pt_flushing.bits, min_d);
		}

		PRCHECK();

		/*
		 * Pass over all the buffers, making sure that each buffer
		 * is clean/invalidated or flushed back to disk marked for
		 * possible invalidation.
		 */
		ebp = &prbufs[prnbufs];
		for (bp = prbufs; bp < ebp; bp++) {
			if ((bp->b_flags & B_INVAL) != 0) {
				/* bp is on the inval list or soon will be */
				continue;
			}

			if ((bp->b_flags & B_ERROR) != 0 && passno > 0) {
				/*
				 * This buffer has had an error.  If this
				 * isn't our first time scanning the list,
				 * then just return with the error to
				 * prevent from scanning the list forever.
				 */
				/* ASSERT(prstatus.pr_status == PRERROR); */
				continue;
			}

			if (dev == NODEV) {
				maj_d = major(bp->b_dev);
				min_d = minor(bp->b_dev);
				ASSERT(PR_MAJORBDEV_OK(maj_d));
				if (isset(prtabs[maj_d]->pt_error.bits,
				    min_d)) {
					/*
					 * This device hasn't been successfully
					 * opened yet, IO isn't allowed yet.
					 */
					continue;
				}
			} else if (dev != bp->b_dev) {
				/* wrong device */
				continue;
			}

			if ((bp->b_flags & (B_DELWRI | B_BUSY)) == 0) {
				/* bp is on the clean list */
				if (inval) {
					/* mark the buffer invalid */
					prbremfree(bp);
					bp->b_flags |= B_INVAL;
					prbrelse(bp);
				}
				continue;
			}

			if ((bp->b_flags & B_ACTIVE) != 0) {
				/*
				 * bp is currently being written back,
				 * mark it for invalidation when it is being
				 * reaped if we were called for invalidation.
				 */
				if (inval) {
					bp->b_flags |= B_INVAL;
				}
				continue;
			}

			if ((bp->b_flags & (B_DELWRI | B_BUSY)) == B_DELWRI) {
				/*
				 * bp is on the dirty list, write
				 * it back and possibly invalidate.
				 */
				PRDEBUG1(">", 0, 0, 0, 0);
				prbremfree(bp);
				if (inval) {
					bp->b_flags |= B_INVAL;
				}
				prbwrite(bp);
			} else {
				/*
				 * Otherwise, it is just being held by
				 * someone and we have to wait on it.
				 */
				while ((bp->b_flags & B_BUSY) != 0) {
					PRDEBUG1(
					"prflushdev: wait bp %x, flags %x\n",
					    bp, bp->b_flags, 0, 0);
					bp->b_flags |= B_WANTED;
					(void) sleep((caddr_t)bp, PRIBIO);
				}
				/*
				 * backup so we look at this bp again
				 */
				bp--;
			}

			/*
			 * Check to see of somebody else
			 * finished flushing before we did.
			 * If so, then reset our write lock
			 * on the cache (and scan_needed
			 * will cause another pass).
			 */
			scan_needed = maybe_slept = 1;
			if (dev != NODEV) {
				setbit(prtab->pt_flushing.bits, min_d);
			} else if (prreadonly == 0) {
				prreadonly = 1;
			}
		}

		/*
		 * Now wait for the active list to go back to zero.
		 */
		while (prstatus.pr_nactive > 0) {
			PRDEBUG1("[%d] ", prstatus.pr_nactive, 0, 0, 0);
			prbwait(*practive);
			scan_needed = maybe_slept = 1;
			if (dev != NODEV) {
				setbit(prtab->pt_flushing.bits, min_d);
			} else if (prreadonly == 0) {
				prreadonly = 1;
			}
		}
		if (*practive != NULL) {
			panic("prflushdev");
			/* NOTREACHED */
		}
	}

	if (dev != NODEV) {
		clrbit(prtab->pt_flushing.bits, min_d);
	} else {
		prreadonly = prstatus.pr_state != PRUP;
	}

	/*
	 * Note - on real disk errors below us, the buffers that
	 * cannot be written out will remain on the dirty list with
	 * an error flag set.
	 */
	PRCHECK();

	return (maybe_slept);
}

/*
 * Write all data from the presto buffers, and if all the writes were
 * successful, reinitialize the presto buffers with the current maxsize.
 *
 * This routine is called any time we have a context (i.e. not actively
 * rebooting), and we have been rebooted with dirty data in the presto
 * cache, and not all the opens for the various devices which the dirty
 * buffers reside have been done.
 *
 * We will stay in the resetting sub state of PRERROR until we reap the last
 * buffer cleanly.
 */
static void
prreset()
{
	register struct buf *ebp, *bp;

	if (prresetting == 0) {
		prresetting = 1;
		printf("presto: writing dirty buffers\n");
	}

	/*
	 * Call the open routine for each device we are about to do IO on.
	 *
	 * Model for the prtab per device bits when
	 *  ((prstatus.pr_state == PRERROR) && (prresetting || prresetneeded))
	 *
	 *   error   enabled
	 *   _____   _______
	 *    0		0	-- open not yet attempted.
	 *    0		1	-- open successfully completed.
	 *    1		1	-- open/verify failed or is in-progress.
	 *    1		0	-- shouldn't happen.
	 */
	ebp = &prbufs[prnbufs];
	for (bp = prbufs; bp < ebp; bp++) {
		if ((bp->b_flags & B_DELWRI) != 0) {
			register int min_d, maj_d;
			register struct prtab *prtab;

			maj_d = major(bp->b_dev);
			min_d = minor(bp->b_dev);
			prtab = prtabs[maj_d];

			if (isclr(prtab->pt_error.bits, min_d) &&
			    isclr(prtab->pt_enabled.bits, min_d)) {
				register int err;

				setbit(prtab->pt_enabled.bits, min_d);
				setbit(prtab->pt_error.bits, min_d);
				/*
				 * Verify that we can open and use the device,
				 * possibly turning on bounceio if needed
				 */
				err = prverifydev(bp->b_dev,
				    prtab->pt_strategy);
				if (err) {
					printf(
"presto: skipping writing of block device (%d, %d) buffers...\n", maj_d, min_d);
				} else {
					/*
					 * Get those puppies out!
					 */
					clrbit(prtab->pt_error.bits, min_d);
					(void) prflushdev(bp->b_dev, TRUE);
					/*
					 * prreap() will have called
					 * prreset_finish() if this was
					 * the last device to be opened
					 * and all the buffers got out ok.
					 */
				}
			}
		}
	}
	prresetneeded = 0;		/* don't call prreset() anymore */
}

/*
 * Get out of this PRERROR "resetting" sub state.
 *
 * All the opens of devices which we had buffers for worked, and all those
 * buffers were succesfully written to disk.
 */
static void
prreset_finish()
{
	register struct prtab **prtabp;

	/*
	 * Now turn off any of the enabled bits that were
	 * set during the "opening/verification" phase.
	 */
	for (prtabp = prtabs; prtabp < &prtabs[nprbdev]; prtabp++) {
		if (*prtabp == NULL)		/* a sparse array */
			continue;
		bzero((char *)&(*prtabp)->pt_enabled,
		    sizeof ((*prtabp)->pt_enabled));
	}

	/*
	 * Model for the prtab bits from now on.
	 *
	 *  error    enabled
	 *  _____    _______
	 *    0		0	-- device not enabled
	 *    0		1	-- device enabled
	 *    1		X	-- device opened failed, no writebacks allowed
	 */
	prresetting = 0;
	prchksum_ignore = 0;
	prnvinit(prstatus.pr_maxsize);
}

/*
 * Ask the user about tossing or writing dirty buffers.
 * Returns TRUE if user said to proceed with write.
 * Returns FALSE and initializes NV and presto memory as well
 * as resetting prresetneeded if user said to toss the dirty buffers.
 */
static int
ask_user(halt_reason)
	char *halt_reason;
{
	char buf[256];		/* big enough hopefully */

	printf("\nYou have three options:\n\
	1) Throw away all dirty buffers.\n\
	2) Write all dirty buffers to disk.\n\
	3) Halt the machine%s.\n", halt_reason);
	for (;;) {
		printf("Press the 1, 2, or 3 key, followed by <RETURN>: ");
		gets(buf);
		switch (*buf) {
			case '1':
				/*
				 * Throwing the data away...
				 */
				prinit(prstatus.pr_maxsize);
				prnvinit(prstatus.pr_maxsize);
				prresetneeded = FALSE;
				return (FALSE);

			case '2':
				/*
				 * The user said to continue on...
				 */
				return (TRUE);

			case '3':
				/*
				 * Wait for power down.
				 */
				printf("Power system down now.\n");
				for (;;)
					/* nothing */;
				/* NOTREACHED */

			default:
				break;
		}
		/* NOTREACHED */
	}
	/* NOTREACHED */
}


/*
 * Handle all the work needed after a reboot.
 *
 * For each potentially dirty buffer
 * do
 *	Copy back bp info from NV memory to corresponding presto buffer
 *	Hash in the presto buffer
 * done
 *
 * If dirty buffers where found, then verify that we can handle the
 * write backs and setup things so they get flushed at the first
 * opportunity (once we have a context).
 */
static void
prreboot(nvh)
	struct nvh *nvh;
{
	register struct buf *bp, *dp;
	struct nvbuf nvbuf, *nvbp;
	register u_int nbp;
	register int i;
	u_long orig_prdirty_led = 0;
	int nbadchksum = 0;

	if (prdirty_led != NULL) {
		/*
		 * Save the original LED dirty state before
		 * it is reset when we find a dirty buffer
		 * and release it through prbrelse().
		 */
		orig_prdirty_led = *prdirty_led;
	}

	nbp = nvh->nvh_nbufs;
	if (prnbufs < nbp) {
		panic("prreboot");
		/* NOTREACHED */
	}
	nvbp = &nvhp->nv_bufs[0];
	bp = prbufs;
	for (i = 0; i < nbp; i++, nvbp++, bp++) {
		BUSREAD((caddr_t)nvbp, (caddr_t)&nvbuf, sizeof (nvbuf));
		if ((nvbuf.nb_flags & (B_DELWRI|B_ACTIVE)) != 0) {
			/*
			 * A dirty buffer or active buffer,
			 * insert it into our buffer list.
			 * As soon as we have a context, we
			 * will start the io.
			 */
			prbremfree(bp);
			bp->b_flags = B_DELWRI | B_BUSY;
#ifdef USE_B_CHAIN
			bp->b_chain = 0;
#else
			proldflags[bp - prbufs] = 0;	/* force write back */
#endif
			bp->b_bcount = nvbuf.nb_bcount;
			bp->b_dev = nvbuf.nb_dev;
			bp->b_blkno = nvbuf.nb_blkno;

			dp = PRBUFHASH(bp->b_dev, bp->b_blkno);
			binshash(bp, dp);

			prbrelse(bp);		/* put back on dirty queue */

			/*
			 * Move into a "read-only" caching state,
			 * and request a reset on flush success.
			 */
			prresetneeded = TRUE;

			/*
			 * If we are using the same checksum algorithm,
			 * then check the buffer now to make sure there was
			 * no bit rot during the time the system was down.
			 */
			if (nvh->nvh_chksum == CHKSUM_VERS &&
			    !prverifybuf(bp->b_un.b_addr, bp->b_bcount))
				nbadchksum++;
		}
	}

	if (prresetneeded) {
		printf("presto: %d dirty buffer%s found\n",
		    prstatus.pr_ndirty, prstatus.pr_ndirty == 1 ? "" : "s");
	}

	if (prresetneeded && nbadchksum != 0) {
		/*
		 * We had some buffers with bad checksums.
		 *
		 * Ask the user what they want us to do here...
		 */
		printf(
"presto: %d dirty buffer%s appear%s to have a bad check sum\n", nbadchksum,
		    nbadchksum == 1 ? "" : "s",
		    nbadchksum == 1 ? "s" : "");
		if (ask_user("")) {
			/*
			 * Set up to ignore checksum errors
			 * during initial writebacks.
			 */
			prchksum_ignore = TRUE;
		}
	}

	if (prresetneeded && nvh->nvh_machineid != machineid()) {
		/*
		 * This board was last running in a different system,
		 * and has data left over from the previous system.
		 *
		 * Ask the user what they want us to do here...
		 */
		printf(
"presto: Dirty data found for host id 0x%x,\n\
which is different from this host's id (0x%x)\n\
Prestoserve contains data which has not been written to any disk yet.\n",
		    nvh->nvh_machineid, machineid());
		if (ask_user(
		    " (to return Prestoserve board to its proper host)")) {
			/*
			 * Update the stored host id info.
			 */
			nvh->nvh_machineid = machineid();
			BUSWRITE((caddr_t)&nvhp->nv_nvh, (caddr_t)&nvh,
			    sizeof (struct nvh));
		}
	}

	if (prresetneeded && prdirty_led != NULL &&
	    (orig_prdirty_led & 0x1) == 0) {
		/*
		 * This board looked like it has dirty data,
		 * but the original dirty LED state said it didn't.
		 *
		 * Ask the user what they want us to do here...
		 */
		*prdirty_led = orig_prdirty_led;
		printf(
"presto: Prestoserve board contains inconsistent state.\n\
It appears to have data which has not been written to any disk yet,\n\
but the dirty data LED claims that Prestoserve has no dirty data.\n");
		if (ask_user("")) {
			/*
			 * Reset the LED back to mark as "dirty".
			 */
			*prdirty_led = TRUE;
		}
	}

	if (prresetneeded) {
		prreadonly = TRUE;
		prstatus.pr_state = PRERROR;
		if (nvh->nvh_chksum != CHKSUM_VERS)
			prchksum_ignore = TRUE;
	} else {
		prnvinit(prstatus.pr_maxsize);
	}

	/*
	 * Now that we are done with host and data checking,
	 * make sure that the kernel has a Presto-ized device
	 * present for each buffer in the dirty list.
	 */
	dp = (struct buf *)&prbfreelist[PR_DIRTY];
	for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
		if (!PR_MAJORBDEV_OK(major(bp->b_dev))) {
			printf(
"presto: Block device %d not present or is not Presto-ized in this kernel!\n",
			    major(bp->b_dev));
			panic("presto startup");
			/* NOTREACHED */
		}
	}

	PRCHECK();
}

/*
 * Do a hashed lookup to see if we can find a buffer containing
 * the given <dev, blkno, size>.  If a partial match or a busy
 * match is found, return failure.
 */
static struct buf *
prbuflookup(dev, blkno, size)
	dev_t dev;
	register daddr_t blkno;
	int size;
{
	register struct buf *bp, *dp;
	register daddr_t lastreq, lastblk;

	PRCHECK();
	lastreq = blkno + btodb(size) - 1;

	/*
	 * N.B. - any overlap conflicts
	 * will be on the same hash chain.
	 */
	dp = PRBUFHASH(dev, blkno);
	for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
		lastblk = bp->b_blkno + btodb(bp->b_bcount) - 1;
		if (lastblk < blkno || bp->b_blkno > lastreq ||
		    bp->b_dev != dev)
			continue;

		PRCHECK();
		return (bp);
	}
	return ((struct buf *)NULL);
}
/*
 * Do a hashed lookup to see if we can find a buffer containing
 * the given <dev, blkno, size>.  We do not return a bp on partial
 * overlapping requests.  We will flush back and invalidate any
 * overlapping buffers found.
 */
static struct buf *
prbufsearch(dev, blkno, size)
	dev_t dev;
	register daddr_t blkno;
	int size;
{
	register struct buf *bp, *dp;
	register daddr_t lastreq, lastblk;

	/*
	 * Cannot have requests cross PRBSIZE boundaries.
	 */
	if ((dbtob(blkno) / PRBSIZE) != ((dbtob(blkno) + size - 1) / PRBSIZE)) {
		panic("prbufsearch");
		/* NOTREACHED */
	}

	PRCHECK();
	lastreq = blkno + btodb(size) - 1;

	/*
	 * N.B. - any overlap conflicts
	 * will be on the same hash chain.
	 */
	dp = PRBUFHASH(dev, blkno);
loop:
	for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
		lastblk = bp->b_blkno + btodb(bp->b_bcount) - 1;
		if (lastblk < blkno || bp->b_blkno > lastreq ||
		    bp->b_dev != dev)
			continue;

		/*
		 * Either a hit or an partial overlap,
		 * wait for the buffer if needed and then
		 * pull the presto buffer off the free list.
		 */
		if ((bp->b_flags & B_BUSY) != 0) {
			prbwait(bp);
			goto loop;		/* start over */
		}
		prbremfree(bp);

		if (blkno >= bp->b_blkno && lastreq <= lastblk) {
			/* Got a real hit */
			PRCHECK();
			return (bp);
		} else {
			/*
			 * Partial overlap - invalidate the buffer.
			 */
			bp->b_flags |= B_INVAL;
			if ((bp->b_flags & B_DELWRI) != 0) {
				/*
				 * Dirty - try to flush back.
				 */
				prbwrite(bp);
				prbwait(bp);	/* wait for IO to finish */
				if ((bp->b_flags & B_ERROR) == 0)
					goto loop;	/* start all over */
				/* else, we had an error - give up and go on */
			} else {
				prbrelse(bp);	/* release and inval the buf */
				goto loop;	/* start all over */
			}
		}
	}
	return ((struct buf *)NULL);
}

static struct diskhd dirtylist = {
	B_HEAD,
	NULL,
	NULL,
	(struct buf *)&dirtylist,
	(struct buf *)&dirtylist,
	0
};

/*
 * Write back the oldest nwrite dirty presto buffers.
 * We batch the write back requests to try and get
 * better utilization of disk arm movement.
 */
static void
prwritedirty(nwrite)
	int nwrite;
{
	register struct buf *dp, *bp;
	register int nwr;

	dp = (struct buf *)&prbfreelist[PR_DIRTY];
	bp = dp->av_forw;
	if (nwrite <= 0) {
		nwrite = 1;		/* at least do one write */
	}

	for (nwr = 0; nwr < nwrite && bp != dp; nwr++) {
		bremfree(bp);
		bp->b_flags |= B_BUSY;
		prstatus.pr_ndirty--;
		binstailfree(bp, (struct buf *)&dirtylist);
		/*
		 * look for contiguous buffers, so we can write them too...
		 */
		if (prcontigbufs) {
			register int i = prcontigbufs;
			register int j;

			do {
				/*
				 * for prcontigbufs buffers...
				 */
				register int blkno;

				j = prcontigspan;
				blkno = bp->b_blkno + btodb(PRBSIZE);
				do {
					/*
					 * look forward upto prcontigspan
					 * blocks for a buffer to write
					 */
					register struct buf *bp2;

#ifdef PRDEBUG
					prlookedat++;
#endif /* PRDEBUG */
					if ((bp2 = prbuflookup(bp->b_dev,
					    blkno, PRBSIZE)) != NULL &&
					    (bp2->b_flags == B_DELWRI) ) {
						bp = bp2;
#ifdef PRDEBUG
						prcontigswritten[prcontigspan - j]++;
#endif /* PRDEBUG */
						bremfree(bp);
						bp->b_flags |= B_BUSY;
						prstatus.pr_ndirty--;
						binstailfree(bp, (struct buf *)&dirtylist);
						PRTRACE(".", 0, 0);
						i--;
						break;
					}
					blkno += btodb(PRBSIZE);
					j--;
				} while (j > 0);
			} while (j > 0 && i > 0);
		}
		bp = dp->av_forw;
	}

	PRTRACE(" {", 0, 0);
	dp = (struct buf *)&dirtylist;
	while ((bp = dp->av_forw) != dp) {
		bremfree(bp);
		prbwrite(bp);
	}
	PRTRACE("} ", 0, 0);
}

/*
 * Flush back any buffers in the range of (off..off+len]
 * If flags & B_INVAL, invalidate all the buffers as well.
 */
static void
prflush(dev, off, len, flag)
	dev_t dev;
	int off, len;
	int flag;
{
	register daddr_t blk, eblk, lastblk;
	register struct buf *bp, *dp;

	PRCHECK();

	if (prresetneeded)
		prreset();

	lastblk = btodb(off + len + DEV_BSIZE - 1);
	/*
	 * Loop over each PRBSIZE chunk.
	 */
loop:
	for (blk = btodb(off); blk < lastblk;
	    blk = (blk + btodb(PRBSIZE)) & ~(btodb(PRBSIZE) - 1)) {
		eblk = min((u_int)lastblk,
		    (u_int)(blk + btodb(PRBSIZE)) & ~(btodb(PRBSIZE) - 1));
		dp = PRBUFHASH(dev, blk);
		/*
		 * Look for buf's from (blk..eblk) in this hash chain.
		 * Note that the hash function computes the
		 * same hash function for block numbers within
		 * the same btodb(PRBSIZE) chunk.
		 */
		for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
			if (bp->b_blkno + btodb(bp->b_bcount) - 1 < blk ||
			    bp->b_blkno > eblk || bp->b_dev != dev ||
			    (bp->b_flags & (B_ERROR | B_INVAL)) != 0)
				continue;
			if ((bp->b_flags & B_BUSY) != 0) {
				prbwait(bp);
				goto loop;	/* start over */
			}
			if ((bp->b_flags & B_DELWRI) == 0 && flag == 0) {
				/*
				 * Buffer is clean and we are not
				 * invalidating, nothing to do.
				 */
				continue;
			}
			prbremfree(bp);
			bp->b_flags |= flag;	/* if B_INVAL, inval's bp */
			if ((bp->b_flags & B_DELWRI) != 0) {
				/*
				 * Dirty - need to flush back
				 */
				PRDEBUG3("prflush: writing %x blkno %x\n",
				    bp, bp->b_blkno, 0, 0);
				prbwrite(bp);
				prbwait(bp);	/* wait for IO to finish */
			} else {
				PRDEBUG3("prflush: invalidating %x blkno %x\n",
				    bp, bp->b_blkno, 0, 0);
				prbrelse(bp);	/* release and inval the buf */
			}
			goto loop;		/* start over */
		}
	}
	PRCHECK();
}

/*
 * Allocate a presto buffer for the given parameters.
 * Returns NULL if buffer is not allocated because of
 * we had to sleep and we have to start all over or
 * because of a presto error condition.
 */
static struct buf *
prbufalloc(dev, blkno, size)
	dev_t dev;
	daddr_t blkno;
	int size;
{
	register struct buf *bp, *dp;
	int nwrite;

	PRCHECK();

	for (;;) {
		ASSERT(prstatus.pr_state == PRUP);

		/* First check for invalidate buffers */
		if (prstatus.pr_ninval > 0) {
			dp = (struct buf *)&prbfreelist[PR_INVAL];
			break;
		}

		/* Next try the clean list */
		if (prstatus.pr_nclean > 0) {
			dp = (struct buf *)&prbfreelist[PR_CLEAN];
			break;
		}

		/* Next try reaping active buffers */
		if (prstatus.pr_nactive > 0) {
			int c;

			c = prstatus.pr_nactive;
			if (prreap() || prstatus.pr_state != PRUP) {
				/*
				 * We could have slept during reap,
				 * or an error was noticed, so we
				 * have to give up and return NULL.
				 */
				PRDEBUG1("prbufalloc: prreap slept\n",
				    0, 0, 0, 0);
				return ((struct buf *)NULL);
			}
			if (c != prstatus.pr_nactive) {
				/*
				 * If the number of active buffers
				 * has changed, rescan the free lists.
				 */
				continue;
			}
		}

		nwrite = (prstatus.pr_ndirty * prwrite_percent) / 100;
		PRDEBUG2("prbufalloc: writing %d bufs\n", nwrite, 0, 0, 0);
		prwritedirty(nwrite);

		if (*practive) {
			/*
			 * We sleep on the first buffer in the to be cleaned
			 * list in the hope that it is the first IO completed.
			 * Finally we return NULL to indicate error condition.
			 */
			prbwait(*practive);
			PRDEBUG1("prbufalloc: waited for IO\n", 0, 0, 0, 0);
		}
		return ((struct buf *)NULL);
	}

	/*
	 * Pull the bp off of the free list given by dp
	 * and destroy the old identity, invalidate any
	 * conflicts, and establish the new identity.
	 */
	bp = dp->av_forw;
	if (dp == bp || (bp->b_flags & (B_BUSY | B_ACTIVE)) != 0) {
		panic("prbufalloc");
		/* NOTREACHED */
	}

	/*
	 * Unhash this buffer, remove it from the free list,
	 * and set up bp with its new identity.
	 */
	if ((bp->b_flags & B_INVAL) == 0) {
		bremhash(bp);
	}
	prbremfree(bp);

	bp->b_flags = B_BUSY;
	dp = PRBUFHASH(dev, blkno);
	binshash(bp, dp);
	bp->b_dev = dev;
	bp->b_error = 0;
	bp->b_blkno = blkno;
	bp->b_bcount = size;
#ifdef USE_B_CHAIN
	bp->b_chain = NULL;
#else
	proldflags[bp - prbufs] = 0;
#endif
	bptonvbp(bp);

	PRCHECK();

	return (bp);
}

/*
 * Returns the number of bytes successfully verified for use with presto.
 */
static u_int
doprobe(reg, maxsize)
	caddr_t reg;
	u_int maxsize;
{
	register u_int size;
	int c;
	long *lp, lv, lv2;

	if (((int)reg & PAGEOFFSET) != 0) {
		printf("pr0: presto memory is not on a page boundary\n");
		return (0);
	}

	/*
	 * After a sample read probe, verify that the data path will
	 * work correctly by reading and writing some long bit patterns.
	 * By doing this first, we prevent the sizing operation from trashing
	 * Presto's memory if it reads back garbage, but it can really write.
	 * Pick an address in the middle of nv_scratch part of the header that
	 * is save to trash on either an mm6704 or an mm1800 board.
	 */

	c = (int)&(((struct nv *)reg)->nv_scratch[DEV_BSIZE/2]);
	c &= ~(sizeof (long) - 1);	/* insure that it's long word aligned */
	lp = (long *)c;

	if (peekl(lp, &lv) == PEEK_FAULT) {
		/*
		 * Cannot read the scratch area,
		 * the board must not be present.
		 *
		 * If the ps device was found,
		 * issue a warning message now.
		 */
		if (psaddr != NULL) {
			printf(
"ps0: Warning! pr0 was not found, so Prestoserve is not operational.\n");
			printf(
"Reseat Prestoserve board and/or check hardware/software address settings.\n");
		}
		return (0);
	}

#define	BITPATTERN	0xaaaaaaaa	/* alternate 1's and 0's */
	lv = BITPATTERN;
	if (pokel(lp, lv) || peekl(lp, &lv) == PEEK_FAULT ||
	    lv != BITPATTERN) {
		printf(
"pr0: failed to write and read 0x%x from board offset 0x%x\n",
		    BITPATTERN, (caddr_t)lp - reg);
		return (0);		/* failed to write and read pat #1 */
	}

	lv = ~BITPATTERN;
	if (pokel(lp, lv) || peekl(lp, &lv) == PEEK_FAULT ||
	    lv != ~BITPATTERN) {
		printf(
"pr0: failed to write and read 0x%x from board offset 0x%x\n",
		    ~BITPATTERN, (caddr_t)lp - reg);
		return (0);		/* failed to write and read pat #2 */
	}
#undef BITPATTERN

	/*
	 * It looks like we can read and write memory ok
	 * (i.e., there appears to be no stuck data lines).
	 * Size the memory by making sure that it all
	 * can be read and written without fail.
	 */
	for (size = 0; size < maxsize; size += PRSTEP) {
		lp = (long *)(reg + size + PRSTEP - sizeof (*lp));
		if (peekl(lp, &lv) == PEEK_FAULT || pokel(lp, lv))
			break;
		/*
		 * Verify that the data actually changes as
		 * expected (the SBus boards will not fault trying
		 * to do reads and writes past the end of memory,
		 * but any data written goes into the bit bucket).
		 */
		if (pokel(lp, ~lv) || peekl(lp, &lv2) == PEEK_FAULT ||
		    pokel(lp, lv) || lv2 != ~lv) {
			/*
			 * We shouldn't have failed doing the peekl()'s
			 * or the pokel()'s as they worked the first time.
			 * Be sure that the original data is put back.
			 */
			(void) pokel(lp, lv);
			break;
		}
	}

	if (size < PRMINSIZE) {
		printf("pr0: only 0x%x bytes found -- too small to use!\n",
		    size);
		return (0);
	}

	/*
	 * If we have already found a ps device,
	 * then we are a mm6704 (Model 500) board.
	 * If we haven't found a ps device, then
	 * we might be a mm1800 (Model 900) board.
	 *
	 * N.B. we depend on the ps before the pr
	 * device in the kernel config file,
	 * presto_chango -i does this for us.
	 */
	if (psaddr == NULL) {
		/*
		 * If we didn't find the ps device, then let's us
		 * see if this might be an mm1800 style board.
		 *
		 * Verify it has the Legato magic number and
		 * that writes to the status registers are ignored.
		 */
		lp = (long *)reg;
		if (peekl(lp, &lv) == PEEK_FAULT ||
		    ((u_char)(lv & PR_MM1800_MAGIC_MASK)) != PR_MM1800_MAGIC ||
		    pokel(lp, ~(long)PR_MM1800_MAGIC) ||
		    (peekl(lp, &lv)) == PEEK_FAULT ||
		    ((u_char)(lv & PR_MM1800_MAGIC_MASK)) != PR_MM1800_MAGIC) {
			return (0);
		}

		prstatus.pr_battcnt = PR_BATTCNT(lv);

		if (prstatus.pr_battcnt < PR_BATTMIN ||
		    prstatus.pr_battcnt > PR_BATTMAX)
			return (0);		/* illegal # of batteries */

		/* The header on 1800 style board starts below the registers */
		nvhp = (struct nv *)(reg + PR_MM1800_MEM_OFFSET);

		/* Save pointer to where the "dirty LED" can be set */
		prdirty_led = (u_long *)(reg + PR_MM1800_LED_OFFSET);

		/* Battery registers are in 5th byte on the 1800 style board. */
		battery_registers = (u_char *)(reg + PR_MM1800_BAT_OFFSET);

		/*
		 * Now perform the psprobe()/psattach()
		 * equivalent for the mm1800.
		 */
		psaddr = (struct ps_statclk *)reg;
		set_battery_mode(BAT_MM1800);
		(void) checkbatteries();
	} else {
		/*
		 * mm6704 board
		 */
		nvhp = (struct nv *)reg;

		/*
		 * The battery register is set in psprobe() and the battery
		 * mode set and initial battery checks are done in psattach().
		 */
	}

	/*
	 * Initialize the checksum array at the
	 * next PRBSIZE boundary beyond the header.
	 *
	 * Putting the checksums in NV RAM means that the data will be
	 * verified after a crash to insure that data was not corrupted.
	 */
	prchksums = (u_int *)PRBALIGN(&nvhp[1]);

	/*
	 * Locate the data buffers at a PRBSIZE boundary
	 * beyond the prchksums array in NV RAM.
	 */
	nvbdp = (struct nvbd *)PRBALIGN(&prchksums[PRMAXSIZE / PRFSIZE]);

	return (size);
}

/*
 * This routine is called to do the real attach work common to
 * boot configuration types.  It is called only at boot time.
 */
static int
doattach(size, unit)
	u_int size;
	int unit;
{
	struct nvh nvh;

	prinit(size);
	BUSREAD(&nvhp->nv_nvh, &nvh, sizeof (nvh));
	if (nvh.nvh_magic == PRMAGIC && !prreinit) {
		if (nvh.nvh_version != PRVERSION) {
			/*
			 * Got a different NV structure version.
			 * Hope that everything was already
			 * flushed, it's too late now...
			 */
			printf("pr%d: expected NV vers %d -- found %d\n",
			    unit, PRVERSION, nvh.nvh_version);
			prnvinit(size);
		} else if (nvh.nvh_bsize != PRBSIZE ||
		    nvh.nvh_fsize != PRFSIZE) {
			/*
			 * Got a different basic or fragment sizes.
			 * Hope that everything was already
			 * flushed, it's too late now...
			 */
			printf("pr%d: expected sizes %d, %d -- found %d, %d\n",
			    unit, PRBSIZE, PRFSIZE,
			    nvh.nvh_bsize, nvh.nvh_fsize);
			prnvinit(size);
		} else if (size == nvh.nvh_size) {
			/*
			 * Normal case, everything looks ok.
			 */
			printf("pr%d: 0x%x bytes\n", unit, size);
			prreboot(&nvh);
		} else if (size > nvh.nvh_size) {
			/*
			 * Found more memory than want we were last using.
			 * Flush everything out and reinitialize everything
			 * to use all of the memory that we just found.
			 */
			printf("pr%d: increased from 0x%x to 0x%x bytes\n",
			    unit, nvh.nvh_size, size);
			prreboot(&nvh);
		} else {
			/*
			 * Wow - we lost some memory this time.
			 * Hope that everything was already
			 * flushed, it's too late now.
			 */
			printf("pr%d: decreased from 0x%x to 0x%x bytes!\n",
			    unit, nvh.nvh_size, size);
			prnvinit(size);
		}
	} else {
		if (prallocmem)
			/* Faking memory */
			printf("pr%d: allocating 0x%x bytes of memory\n",
			    unit, size);
		else
			/* First time through */
			printf("pr%d: initializing 0x%x bytes\n", unit, size);
		prnvinit(size);
	}
	return (0);
}

/*
 * Debugging code - kmem_alloc's the Presto memory space.
 * Returns the size of the kmem_alloc'ed buffer.
 */
static u_int
doallocmem()
{

	if (prallocmem < PRMINSIZE)
		return (0);		/* too small to use */
	prreinit = 1;
	praddr = kmem_alloc(prallocmem);
	nvhp = (struct nv *)praddr;
	psaddr = (struct ps_statclk *)praddr;

	set_battery_mode(BAT_PRETEND);

	/*
	 * Initialize the checksum array at the
	 * next PRBSIZE boundary beyond the header.
	 */
	prchksums = (u_int *)PRBALIGN(&nvhp[1]);

	/*
	 * Locate the data buffers at the next PRBSIZE
	 * boundary beyond the prchksums array.
	 */
	nvbdp = (struct nvbd *)PRBALIGN(&prchksums[PRMAXSIZE / PRFSIZE]);

	prstatus.pr_maxsize = prallocmem;
	return (prallocmem);
}

#ifdef OPENPROMS
/*
 * This routine is called at boot time on SBus/open boot PROM systems.
 *
 * Returns 1 if accepted, 0 if rejected.
 */
static int
prsbusidentify(name)
	char *name;
{
	register char *cp;

	if (praddr != NULL)
		return (0);		/* only one per system for now */

#if defined (SUN_RELEASE) && SUN_RELEASE != 0
	/* machine = (char *)getlongprop(nodeid, "name"); */
	/* kmem_free(machine, strlen(machine)); */

	if (strncmp(top_devinfo->devi_name, "SUNW,", 5) != 0) {
		printf( 
"This version of S-Bus Prestoserve not compatible in this configuration\n");
		return(0);
	}
#endif	/* SUN_RELEASE */

	/*
	 * Skip any leading manufacturing information in
	 * front of the driver name (comma separated).
	 */
	for (cp = name; *cp != '\0' && *cp != ','; cp++)
		/* nothing */;
	if (*cp == ',')
		cp++;
	else
		cp = name;

	/*
	 * See if the driver id string matches our long driver id string
	 */
	if (strcmp(cp, LONG_IDSTRING) != 0 &&
	    !(prallocmem != 0 && strcmp(cp, "options") == 0)) {
		/*
		 * If we are going to kmem_alloc our buffer, accept the
		 * "options" driver so that our attach routine will be
		 * called even when a SBus Prestoserve board isn't present.
		 */
		return (0);
	}

	praddr = (char *)TRUE;		/* temporary value marking a hit */
	return (1);
}

/*
 * SBus/open boot PROM systems attach routine.  Must handle mapping in
 * the memory, verifying that the memory is good, and then all the
 * needed boot time handling for presto state.
 *
 * Returns 0 on success and -1 on failure.
 */
static int
prsbusattach(devinfo_p)
	struct dev_info *devinfo_p;
{
	struct dev_reg *dev_reg_p;
	u_int size;

	if (prallocmem != 0) {
		/*
		 * Just kmem_alloc the requested space.
		 */
		size = doallocmem();
		if (size == 0) {
			praddr = NULL;
			return (-1);
		}
		/*
		 * Now handle all the attach related
		 * work and return the result.  Note
		 * that in this case we don't go thru
		 * the report_dev() stuff at all..
		 */
		prstatus.pr_maxsize = size;
		return (doattach(prstatus.pr_maxsize, 0));
	}

	ASSERT(devinfo_p != NULL);
	dev_reg_p = devinfo_p->devi_reg;
	ASSERT(dev_reg_p != NULL && devinfo_p->devi_nreg == 1);

	praddr = map_regs(dev_reg_p->reg_addr, dev_reg_p->reg_size,
	    dev_reg_p->reg_bustype);

	if (praddr == NULL) {
		printf("pr0: cannot map 0x%x bytes in kernel\n",
		    dev_reg_p->reg_size);
		return (-1);
	}

	size = doprobe(praddr, dev_reg_p->reg_size);
	if (size < dev_reg_p->reg_size)
		unmap_regs(praddr + size, dev_reg_p->reg_size - size);

	if (size == 0) {
		praddr = NULL;
		return (-1);
	}

	if (prmakecachable) {
		register caddr_t va;
		struct pte pte;
#if defined(SUN_RELEASE) && SUN_RELEASE != 0
		UN_PTE pteu;		/* See arch dependent note at top */
		int sun4m;
		char *cp;

		/* Determine what we're running on */
		cp = top_devinfo->devi_name;
		while (*cp != '\0') {
			if (*cp == '4')
				break;
			cp++;
		}

		if ((*cp == '4') & (strncmp(cp, "4/600", 5) == 0))
			sun4m = 1;
		else
			sun4m = 0;
#endif
		/*
		 * Turn off the MMU's no cache bit for the presto memory.
		 * This should only be used on machines where the IO space
		 * is cachable and that have a write-through cache (SS-2).
		 */
		for (va = praddr; va < praddr + size; va += MMU_PAGESIZE) {
			mmu_getpte(va, &pte);

#if defined(SUN_RELEASE) && SUN_RELEASE != 0
			pteu.base = pte;
			if ((sun4m == 1) && (pteu.cache.Cacheable == 0)) {
				if (prmakecachable > 1)
					printf(".");
				pteu.cache.Cacheable = 1;
				mmu_setpte(va, pteu.base);
			}
			if ((sun4m == 0) && (pteu.cache.pg_nc == 1)) {
				if (prmakecachable > 1)
					printf(".");
				pteu.cache.pg_nc = 0;
				mmu_setpte(va, pteu.base);
			}
#else
			if (pte.pg_nc == 1) {
				if (prmakecachable > 1)
					printf(".");
				pte.pg_nc = 0;
				mmu_setpte(va, pte);
			}
#endif
		}
		if (prmakecachable > 1)
			printf("\n");
	}

	/*
	 * Set up to use the short form of the ID string for kernel messages.
	 * Then identify that this board was successfully found and verified.
	 */
	devinfo_p->devi_name = SHORT_IDSTRING;
	devinfo_p->devi_unit = 0;
	report_dev(devinfo_p);

	/*
	 * Now handle all the attach related work and return the result.
	 */
	prstatus.pr_maxsize = size;
	return (doattach(prstatus.pr_maxsize, 0));
}
#endif /* OPENPROMS */

/*
 * Called at boot time to determine if the presto status device is present.
 * If the device is present, then this must be the MM6704 6U board.
 */
/* ARGSUSED */
static int
psprobe(reg, unit)
	caddr_t reg;
	int unit;
{
	register struct ps_statclk *psa;
	int bstatus;

	if (psaddr != NULL) {
		return (0);		/* only one per system */
	}

	psa = (struct ps_statclk *)reg;
	if ((bstatus = peekc((char *)&psa->status)) == PEEK_FAULT) {
		PRDEBUG1("peekc(%x) failed\n", &psa->status, 0, 0, 0);
		return (0);
	}

	/*
	 * Verify that writing the status register
	 * doesn't actually change anything.
	 */
	if (pokec((char *)&psa->status, ~bstatus)) {
		PRDEBUG1("pokec(%x, %x) failed\n", &psa->status, ~bstatus,
		    0, 0);
		return (0);
	}
	if (bstatus != psa->status) {
		PRDEBUG1("bstatus=%x != status=%x\n", bstatus, psa->status,
		    0, 0);
		return (0);
	}

	if ((bstatus & PS_DC) == 0) {
		printf("ps0: status=%b, no +5 volts!\n", bstatus, PS_BITS);
		return (0);
	}

#ifdef COPY_PROTECT
	reg = (caddr_t)&psa->mem_clock;
#ifdef DSW_DEBUG
printf("SMARTwatch @0x%x\n", reg);
#endif
#ifdef PRDEBUG
	if (set_watch) {
		/*
		 * Load with "the magic number" (Don Coleman's birthday)
		 */
		smartwatch.day_of_month_h = 0;
		smartwatch.day_of_month_l = 7;
		smartwatch.month_h = 0;
		smartwatch.month_l = 5;
		smartwatch.year_h = 6;
		smartwatch.year_l = 2;
		smartwatch.osc = 1;	/* keep oscillator turned OFF */

		/*
		 * Write above to the SMARTwatch registers.
		 */
		dsw_write(reg, (unsigned int *)&smartwatch);
	}
#endif /* PRDEBUG */
	dsw_read(reg, (u_int *)&smartwatch);

#ifdef DSW_DEBUG
	/*
	 * So what was it set to?
	 */
	printf("day %d:%d, month %d:%d, year %d:%d, osc %d\n",
	    smartwatch.day_of_month_h, smartwatch.day_of_month_l,
	    smartwatch.month_h, smartwatch.month_l,
	    smartwatch.year_h, smartwatch.year_l, smartwatch.osc);
#endif /* DSW_DEBUG */

	/*
	 * Verify the magic number stored in the SMARTwatch clock.
	 * Note that the printf indicates that someone is trying
	 * to violate copy protect or the SMARTwatch is in trouble.
	 */
	if (smartwatch.day_of_month_h != 0 || smartwatch.day_of_month_l != 7 ||
	    smartwatch.month_h != 0 || smartwatch.month_l != 5 ||
	    smartwatch.year_h != 6 || smartwatch.year_l != 2) {
		printf("ps0: Invalid battery status.\n");
		return (0);
	}
#endif /* COPY_PROTECT */
	psaddr = psa;
	battery_registers = &psa->status;
	return (sizeof (struct ps_statclk));
}

/*
 * This routine is called to initialize the presto status device for
 * the mm6704 after it has been successfully probed.  It is called
 * only at boot time by the auto configuration code.
 */
/* ARGSUSED */
static int
psattach(md)
	struct mb_device *md;
{

	set_battery_mode(BAT_MM6704);
	(void) checkbatteries();
	return (0);
}

/*
 * Called at boot time to determine if the presto memory device is present.
 */
/* ARGSUSED */
static int
prprobe(reg, unit)
	caddr_t reg;
	int unit;
{
	register u_int size;
	u_int rsize;
	int c;

#if defined(SUN_RELEASE) && SUN_RELEASE != 0

#ifndef CPU_SUN4_470
#define SUN4_ARCH       0x20            /* arch value for Sun 4 */
#define MACH_470        0x04
#define CPU_SUN4_470    (SUN4_ARCH + MACH_470)
#endif
#ifndef CPU_SUN4M_690
#define SUN4M_ARCH      0x70            /* arch value for Std referenc mmu */
#define MACH_690        0x01            /* SPARCsystem 600 series */
#define CPU_SUN4M_690   (SUN4M_ARCH + MACH_690)
#endif

	if ((cpu != CPU_SUN4_470) && (cpu != CPU_SUN4M_690)) {
		printf(
"This version of VME Prestoserve operates only with Sun SPARCservers (4XX/6XX)\n");
		return (0);
	}
#endif /* SUN_RELEASE */

	if (praddr != NULL)
		return (0);			/* only one per system */

	if (prallocmem > 0) {
		/*
		 * Free up all the virtual space allocated by auto config code.
		 * We do this now so that the kmem_alloc() is more likely
		 * to succeed.  If the kmem_alloc doesn't work, we won't be
		 * able to fall back to using the real device if it is present.
		 */
		c = btokmx(reg);
		segkmem_mapout(&kseg, reg, (u_int)prdriver.mdr_size);
		rmfree(kernelmap, (long)btoc(prdriver.mdr_size), (u_long)c);

		return (doallocmem());
	}

	size = doprobe(reg, prdriver.mdr_size);
	if (size == 0)
		return (0);

#ifdef sun
	/*
	 * Free up any additional virtual space allocated by auto config code.
	 */
	rsize = ptob(btopr(size));
	if (rsize > 0 && rsize < prdriver.mdr_size) {
		c = btokmx(reg + rsize);
		segkmem_mapout(&kseg, reg + rsize,
		    (u_int)(prdriver.mdr_size - rsize));
		rmfree(kernelmap, (long)btoc(prdriver.mdr_size - rsize), c);
	}
#endif /* sun */

	praddr = reg;
	prstatus.pr_maxsize = size;
	return (size);
}

/*
 * This routine is called to initialize the presto memory device after
 * it has been successfully probed.  It is called only at boot time by
 * the auto configuration code.
 */
static int
prattach(md)
	struct mb_device *md;
{

	if (psaddr == NULL) {
		printf(
"pr0: Warning! ps0 was not found, so Prestoserve is not operational.\n");
		printf(
"Reseat Prestoserve board and/or check hardware/software address settings.\n");
		return (1);
	}

	return (doattach(prstatus.pr_maxsize, md->md_unit));
}

/*
 * Verify that Presto can use this disk.  This includes opening
 * the device and if bounceio is being used for this device,
 * that the device can address the presto memory correctly.
 *
 * If we cannot pass the "directio" test, then we will automatically set
 * up to use "bounceio" mode to avoid all problems.  Unfortunately, with
 * some Sun devices the cost of failure is a kernel crash and thus for
 * many devices (e.g., SCSI on SPARCstation 330's), we must have the
 * device set up for "bounceio" and never attempt to ever do "directio".
 */
int
prverifydev(dev, xxstrategy)
	dev_t dev;
	int (*xxstrategy)();
{
	int maj_d, min_d, err = 0;

	if (praddr == NULL)
		return (ENODEV);

	maj_d = major(dev);
	min_d = minor(dev);

	err = (*bdevsw[maj_d].d_open)(dev, FREAD|FWRITE);
	if (err) {
		printf("presto: error %d opening block device (%d, %d)\n",
		    err, maj_d, min_d);
		err = ENODEV;
	else {
		(void) (*bdevsw[maj_d].d_close)(dev, FREAD|FWRITE);
	}
	return (err);
}

/*
 * Called through the cdevsw open table for generic presto control.
 * This is used to get a handle on the generic presto device so
 * that ioctl's can be done to control all presto indirect devices.
 */
/* ARGSUSED */
int
propen(dev, flag)
	dev_t dev;
	int flag;
{

	if (praddr == NULL || psaddr == NULL)
		return (ENODEV);

	if (prresetneeded)
		prreset();

	return (0);
}

/*
 * Called via the bdevsw stub close table entries for indirect devices.
 * We invalidate any presto buffers for the device (just in case it is
 * a removable device) and then indirect to the appropriate real device
 * close routine (which is normally just a nop).
 */
int
PRclose(dev, flag, xxclose, xxprtab)
	dev_t dev;
	int flag;
	struct prtab *xxprtab;
	int (*xxclose)();
{
	register dev_t min_d = minor(dev);

loop:
	if (prstatus.pr_state == PRERROR) {
		if (isset(xxprtab->pt_error.bits, min_d)) {
			/*
			 * The device came "online" somehow...
			 * If all the buffers successfully get
			 * to disk now, we end up in the PRDOWN state.
			 */
			clrbit(xxprtab->pt_error.bits, min_d);
			(void) prflushdev(dev, TRUE);
			goto loop;
		} else if (prresetneeded) {
			/*
			 * Do initial syncing stuff.
			 */
			prreset();
			goto loop;
		} else if (prresetting ||
		    isset(xxprtab->pt_enabled.bits, min_d)) {
			(void) prflushdev(dev, TRUE);
		}
	} else if (prstatus.pr_state == PRUP &&
	    isset(xxprtab->pt_enabled.bits, min_d)) {
		(void) prflushdev(dev, TRUE);
	}

	/*
	 * We could still be in a PRERROR state,
	 * but we have done our reasonable best.
	 */
	return ((*xxclose)(dev, flag));
}

/*
 * Called thru the bdevsw block device strategy routine for the
 * indirect driver to handle block read and write requests.  The
 * basic idea is to use the presto cache to avoid disk write traffic.
 */
PRstrategy(bp, xxstrategy, xxprtab)
	register struct buf *bp;
	register int (*xxstrategy)();
	struct prtab *xxprtab;
{
	register struct buf *prbp;
	register struct io *iop;
	dev_t bdev = bp->b_dev;
	dev_t minorbdev = minor(bdev);
	int off;
	caddr_t busaddr;

	/*
	 * the following "if" is optimized for the normal UP and DOWN condition.
	 */
loop:
	if (prstatus.pr_state != PRUP) {
		if (prstatus.pr_state == PRDOWN) {
			(*xxstrategy)(bp);
			return;
		} else {
			/*
			 * In the PRERROR state:
			 * All the wierd stuff happens here.
			 */
			if (isset(xxprtab->pt_error.bits, minorbdev)) {
				/*
				 * The device came "online" somehow...
				 * If all the buffers successfully get
				 * to disk now, we end up in the PRDOWN state.
				 */
				clrbit(xxprtab->pt_error.bits, minorbdev);
				(void) prflushdev(bdev, TRUE);
				goto loop;
			} else if (prresetneeded) {
				/*
				 * Do initial syncing stuff.
				 */
				prreset();
				goto loop;
			} else if (! prresetting &&
			    isclr(xxprtab->pt_enabled.bits, minorbdev)) {
				(*xxstrategy)(bp);
				return;
			}
			/* fall through and search for a hit */
		}
	} else if (isclr(xxprtab->pt_enabled.bits, minorbdev)) {
		(*xxstrategy)(bp);
		return;
	} else if (panicstr != NULL) {	/* just noticed a panic! */
		(void) prdown();
		goto loop;
	}

	/*
	 * If we are in the error state, prreadonly must be set.
	 */
	ASSERT(prstatus.pr_state == PRERROR ? prreadonly : 1);

	if ((bp->b_flags & B_READ) != 0)
		iop = &prstatus.pr_rdstats;
	else
		iop = &prstatus.pr_wrstats;
	iop->total++;
	off = dbtob(bp->b_blkno);

	PRDEBUG3("prstrategy: dev=0x%x, blkno=0x%x, count=0x%x ",
	    bp->b_dev, bp->b_blkno, bp->b_bcount, 0);

	if ((off / PRBSIZE) != ((off + bp->b_bcount - 1) / PRBSIZE) ||
	    (bp->b_bcount & (PRFSIZE - 1)) != 0 || (off & (PRFSIZE - 1)) != 0) {
		/*
		 * A IO request that crosses PRBSIZE blocks or a
		 * non-PRFSIZE size or aligned transfer - just punt.
		 *
		 * We need to invalidate any incore caching
		 * when doing a write while we just need to
		 * flush any incore data when doing a read.
		 */
		PRDEBUG2("prstrategy: bad size/alignment off=0x%x, cnt=0x%x\n",
		    off & (PRBSIZE - 1), bp->b_bcount, 0, 0);
		prflush(bdev, dbtob(bp->b_blkno), bp->b_bcount,
		    ((bp->b_flags & B_READ) != 0)? 0 : B_INVAL);

		/* Should pass on error up if the flush failed. XXX */
		(*xxstrategy)(bp);
		iop->pass++;
		return;
	}

	/*
	 * If we are currently below the desired available
	 * presto buffers, write back something now (but
	 * not if we are in the PRERROR state).
	 */
	if (prprewrite && (prstatus.pr_nclean + prstatus.pr_ninval +
	    prstatus.pr_nactive < prdesavail) &&
	    prstatus.pr_state != PRERROR) {
		if (prprewritenum) {
			prwritedirty(prprewritenum);
		} else {
			prwritedirty(prdesavail -
			    (prstatus.pr_nclean + prstatus.pr_ninval));
		}
	}

	if (pr_batt_check_time < time.tv_sec) {
		(void) checkbatteries();
		pr_batt_check_time = time.tv_sec + pr_batt_check_rate;
	}

	/*
	 * Now we know that we have a request that fits in a PRBSIZE'd chunk.
	 * See if we find can a buffer which corresponds to what we need.
	 */
again:
	prbp = prbufsearch(bdev, bp->b_blkno, bp->b_bcount);

	/*
	 * Some general notes on how we handle errors:
	 *
	 * If we are in the error state, then prreadonly is set, and thus
	 * we will not add anything new to our cache, but we will update
	 * what buffers we have with the new data, and try to write it out
	 * again.
	 */

	if (prbp == NULL) {
		/*
		 * We didn't find a matching presto buffer.
		 */
		if ((bp->b_flags & B_READ) == 0) {
			register int async_pass;

			/*
			 * Async requests (which are always locally generated)
			 * are treated specially...
			 */
			if ((bp->b_flags & B_ASYNC) != 0) {
				if (pr_nocache_async) {
					async_pass = 1;
					/*
					 * Update our async cache trigger?
					 */
					if (pr_cache_async_occasionally)
						pr_nocache_async--;
				} else if (prstatus.pr_nclean +
						prstatus.pr_ninval == 0) {
					/*
					 * we have an asynchronous write request
					 * with no matching presto buffer, and
					 * we don't have any free buffers, so
					 * pass on this one so it won't
					 * degrade into a synchronous write when
					 * using presto
					 */
					async_pass = 1;
				} else if (pr_cache_async_occasionally) {
					/*
					 * Update our async cache trigger and
					 * cache this async write, but we won't
					 * cache any more async writes until
					 * pr_nocache_async counts back to 0
					 */
					pr_nocache_async =
						pr_cache_async_occasionally;
					async_pass = 0;
				} else {
					async_pass = 0;
				}
			} else {
				async_pass = 0;
			}

			if (async_pass || prreadonly ||
			    isset(xxprtab->pt_flushing.bits, minorbdev)) {
				/*
				* Ether we are to pass the async write,
				* or nothing new is to be added to the
				* cache, or we are flushing this partition,
				* so just pass the request.
				*/
				(*xxstrategy)(bp);
				iop->pass++;
				PRTRACE("<awpass %x [%x]> ",
				bp->b_blkno, bp->b_bcount);
				return;
			}
			/*
			 * Allocate an appropriate buffer now.
			 */
			prbp = prbufalloc(bdev, bp->b_blkno, bp->b_bcount);
			if (prbp == NULL) {
				goto again;		/* start all over */
			}
			iop->alloc++;
			PRTRACE("<alloc %x> ", bp->b_blkno, 0);
		} else {
			/*
			 * A read request where there is no corresponding
			 * presto buffer.  Just pass it on the real device
			 * without bothering to cache in presto's buffers
			 * since someone above us will probably cache it.
			 */
			(*xxstrategy)(bp);
			iop->pass++;
			PRTRACE("<rmiss %x> ", bp->b_blkno, 0);
			return;
		}
	} else {
		/*
		 * Found an incore presto buffer.
		 */
		if ((prbp->b_flags & B_DELWRI) != 0) {
			iop->hitdirty++;
			PRTRACE("<%cHit %x> ", (bp->b_flags & B_READ) != 0 ?
			    'r' : 'w', bp->b_blkno);
		} else {
			iop->hitclean++;
			PRTRACE("<%chit %x> ", (bp->b_flags & B_READ) != 0 ?
			    'r' : 'w', bp->b_blkno);
		}
	}

	/*
	 * Call bp_mapin() to make the buffer addressable
	 * (at bp->b_un.b_addr) by mapping the buffer into
	 * kernel virtual address space.  On systems other
	 * than SunOS 4.0, this call may have to be written.
	 */
	bp_mapin(bp);

	/*
	 * We now have the presto buffer and an addressable
	 * kernel buffer, just move the data in or out of it
	 * from the right place and we're done.
	 */
	busaddr = prbp->b_un.b_addr + dbtob(bp->b_blkno - prbp->b_blkno);
	if ((bp->b_flags & B_READ) != 0) {
		/*
		 * Read request - copy from presto out to the given bp.
		 */
		prnvbufread(busaddr, bp->b_un.b_addr, (u_int)bp->b_bcount);
	} else {
		/*
		 * Write request - copy from the given bp into the
		 * presto buffer and mark the (entire) buffer as dirty.
		 */
		prnvbufwrite(busaddr, bp->b_un.b_addr, (u_int)bp->b_bcount);
		prbp->b_flags |= B_DELWRI;

		/*
		 * If we had an error on the buffer before, then try to do
		 * the write immediately and see if it works.
		 */
		if (prbp->b_flags & B_ERROR) {
			prbwrite(prbp);
			prbwait(prbp);
			if (prbp->b_flags & B_ERROR) {
				bp->b_flags |= B_ERROR;
				bp->b_error = prbp->b_error;
			}
			bp->b_resid = 0;		/* got all the bytes */
			biodone(bp);			/* mark bp as done */
			return;
		}
	}

	prbrelse(prbp);			/* release the presto buffer */
	bp->b_resid = 0;		/* got all the bytes */
	biodone(bp);			/* mark bp as done */
}

/*
 * The device driver mmap cdev[] function.
 * Returns the proper "pfnum" bits from the pte.
 */
/* ARGSUSED */
int
prmmap(dev, off, prot)
	dev_t dev;
	off_t off;
	int prot;
{

	if (off >= prstatus.pr_maxsize)
		return (-1);
	return ((int)hat_getkpfnum(praddr + off));
}

/*
 * The common handler for device specific reads and writes.
 */
static int
pr_rw_common(cdev, uio, xxfunc, xxprtab, flag)
	dev_t cdev;
	struct uio *uio;
	int (*xxfunc)();
	struct prtab *xxprtab;
	int flag;
{
	register dev_t min_d = minor(cdev);
	dev_t bdev = makedev(xxprtab->pt_bmajordev, min_d);

loop:
	if (prstatus.pr_state == PRERROR) {
		if (isset(xxprtab->pt_error.bits, min_d)) {
			/*
			 * The device came "online" somehow...
			 * If all the buffers successfully get
			 * to disk now, we end up in the PRDOWN state.
			 */
			clrbit(xxprtab->pt_error.bits, min_d);
			(void) prflushdev(bdev, TRUE);
			goto loop;
		} else if (prresetneeded) {
			/*
			 * Do initial syncing stuff.
			 */
			prreset();
			goto loop;
		} else if (prresetting ||
		    isset(xxprtab->pt_enabled.bits, min_d)) {
			prflush(bdev, uio->uio_offset, uio->uio_resid, flag);
		}
	} else if (prstatus.pr_state == PRUP &&
	    isset(xxprtab->pt_enabled.bits, min_d)) {
		/*
		 * For now, just make sure that any dirty buffers over
		 * the range of the raw IO request are flushed back to
		 * disk to avoid stale data problems.
		 */
		prflush(bdev, uio->uio_offset, uio->uio_resid, flag);

		/* if we just had an error, we should so something here XXX */
	}

	return ((*xxfunc)(cdev, uio));
}

/*
 * PRread and PRwrite are called thru the XXread and XXwrite routines
 */
int
PRread(dev, uio, xxread, xxprtab)
	dev_t dev;
	struct uio *uio;
	int (*xxread)();
	struct prtab *xxprtab;
{

	return (pr_rw_common(dev, uio, xxread, xxprtab, 0));
}

int
PRwrite(dev, uio, xxwrite, xxprtab)
	dev_t dev;
	struct uio *uio;
	int (*xxwrite)();
	register struct prtab *xxprtab;
{

	return (pr_rw_common(dev, uio, xxwrite, xxprtab, B_INVAL));
}

/*
 * Attempt to take presto to the PRDOWN state.
 * On return, presto will be in either the PRDOWN
 * state (returning 0) or PRERROR state (returning EBUSY).
 * This routine is used internally when taking presto
 * down and externally by PRreboot() when halting the OS.
 */
int
prdown()
{

	if (prstatus.pr_state == PRDOWN)
		return (0);

	/*
	 * Flush and invalidate everything here to get valid presto status.
	 */
	(void) prflushdev(NODEV, TRUE);
	if (prstatus.pr_state == PRERROR) {
		/*
		 * Not totally successful -- set things up so
		 * when presto comes on line it'll be down.
		 */
		laststate = PRDOWN;
		return (EBUSY);
	}

	/*
	 * Ok, every thing got out, so now shut things down.
	 */
	prstatus.pr_state = PRDOWN;
	pr_went_down = (u_long)time.tv_sec;
	return (0);
}

#ifdef PRCOMPAT
/*
 * Translate a prtab structure to the old style prtab structure
 * for backward compatibility with old Prestoserve releases.
 */
static void
mkoprtab(oprtab, prtab)
	register struct oprtab *oprtab;
	register struct prtab *prtab;
{

	oprtab->strategy = prtab->pt_strategy;
	oprtab->bmajordev = prtab->pt_bmajordev;
	oprtab->flags = BOUNCEIO(prtab, makedev(prtab->pt_bmajordev, 0));
	oprtab->enabled = prtab->pt_enabled;		/* C structure copy */
	oprtab->flushing = prtab->pt_flushing;		/* C structure copy */
	oprtab->error = prtab->pt_error;		/* C structure copy */
}
#endif /* PRCOMPAT */

/*
 * Translate a kernel prtab structure to a user prtab structure.
 */
static void
mkuprtab(uprtab, prtab)
	register struct uprtab *uprtab;
	register struct prtab *prtab;
{

	uprtab->upt_bmajordev = prtab->pt_bmajordev;
	uprtab->upt_enabled = prtab->pt_enabled;	/* C structure copy */
	if ((prflags & PR_BOUNCEIO) == 0) {
		/* C structure copy */
		uprtab->upt_bounceio = prtab->pt_bounceio;
	} else {
		/* turn on all the bits - bounceio is enabled for all */
		register int i;

		for (i = 0; i < sizeof (uprtab->upt_bounceio.bits); i++)
			uprtab->upt_bounceio.bits[i] = 0xff;
	}
}

/*
 * This routine is called through the cdevsw[]
 * ioctl interface for the presto control device.
 */
/* ARGSUSED */
int
prioctl(dev, cmd, data, flag)
	dev_t dev;
	register int cmd;
	register caddr_t data;
	int flag;
{
	prstates state;
	int msize;
	struct presto_status *statusp;

	switch (cmd) {

	case PRSETSTATE:
		state = *(prstates *)data;
		switch (state) {
		case PRDOWN:
			if (!suser())
				return (EPERM);

			return (prdown());

		case PRUP:
			if (prstatus.pr_state == PRUP)
				return (0);

			/*
			 * Anybody is allowed to ask presto to come
			 * "back up" if it is currently in the error state.
			 */
			if (prstatus.pr_state == PRERROR) {
				/*
				 * Try flushing and invalidating things
				 * and see if things are now ok.
				 */
				(void) prflushdev(NODEV, TRUE);
				if (prstatus.pr_state == PRERROR) {
					if (prstatus.pr_ndirty != 0 ||
					    checkbatteries() != 0) {
						return (EBUSY);
					}
				}
			} else if (!suser()) {
				return (EPERM);
			}

			clearstats();
			prstatus.pr_state = PRUP;
			pr_went_up = (u_long)time.tv_sec;
			prreadonly = 0;	/* allow writes to the cache */
			return (0);

		case PRERROR:
		default:
			return (EINVAL);
		}

	case PRSETMEMSZ:
		if (!suser())
			return (EPERM);
		msize = *(int *)data;

		if (msize < PRMINSIZE)
			return (EINVAL);

		if (msize > prstatus.pr_maxsize)
			return (ENOMEM);

		/*
		 * Get out of the reseting state if we are still there
		 */
		if (prresetting) {
			prresetneeded = 0;
			prreset_finish();
		}

		/*
		 * PRSETMEMSZ ioctl retains the PRUP state.
		 */
		state = prstatus.pr_state;

		if (prdown())
			return (EBUSY);		/* can't flush all buffers */

		prinit(msize);
		prnvinit(msize);
		prstatus.pr_state = (state == PRUP)? PRUP : PRDOWN;

		/*
		 * Reset the batterystatus variable to the normal
		 * value.  If the batterystatus is not currently
		 * in the normal condition, the checkbatteries()
		 * will print the current battery status and disable
		 * presto if all the batteries are currently low.
		 */
		batterystatus = Pr_normal;
		(void) checkbatteries();
		return (0);

	case PRGETSTATUS:
		statusp = (struct presto_status *)data;
		(void) prreap();		/* reap any presto bufs */
		(void) checkbatteries();	/* get battery status */
		*statusp = prstatus;		/* structure copy */
		if (prstatus.pr_state == PRDOWN) {
			if (pr_went_down <= pr_stats_cleared)
				statusp->pr_seconds = 0;
			else
				statusp->pr_seconds =
					pr_went_down - pr_stats_cleared;
		} else if (pr_stats_cleared) {
			statusp->pr_seconds =
				(u_long)time.tv_sec - pr_stats_cleared;
		} else {
			statusp->pr_seconds = 0;
		}

		return (0);

	case PRRESET:		/* reset the entire presto state */
		if (!suser())
			return (EPERM);

		/*
		 * Get out of the reseting state if we are still there
		 */
		if (prresetting) {
			prresetneeded = 0;
			prreset_finish();
		}

		/*
		 * Try to shutdown things cleanly before reinitializing
		 */
		(void) prdown();
		prstatus.pr_state = PRDOWN;

		prinit(prstatus.pr_cursize);
		prnvinit(prstatus.pr_cursize);

		/*
		 * Reset the batterystatus variable to the normal
		 * value.  If the batterystatus is not currently
		 * in the normal condition, the checkbatteries()
		 * will print the current battery status and disable
		 * presto if all the batteries are currently low.
		 */
		batterystatus = Pr_normal;
		(void) checkbatteries();
		return (0);

	case PRENABLE:		/* enable presto buffering on a block dev */
		{
			register dev_t bdev = *((dev_t *)data);
			register dev_t maj_d = major(bdev);
			register dev_t min_d = minor(bdev);
			register int err;
			register struct prtab *prtab;

			if (!PR_MAJORBDEV_OK(maj_d))
				return (ENODEV);

			if (!suser())
				return (EPERM);

			if (prstatus.pr_state == PRERROR)
				return (EBUSY);

			prtab = prtabs[maj_d];
			err = prverifydev(bdev, prtab->pt_strategy);
			if (err != 0)
				return (err);
			setbit(prtab->pt_enabled.bits, min_d);
		}
		return (0);

	case PRDISABLE:		/* disable presto buffering on a block dev */
		{
			register dev_t bdev = *((dev_t *)data);
			register dev_t maj_d = major(bdev);
			register dev_t min_d = minor(bdev);

			if (!PR_MAJORBDEV_OK(maj_d))
				return (ENODEV);

			if (!suser())
				return (EPERM);

			(void) prflushdev(bdev, TRUE);

			if (prstatus.pr_state == PRERROR)
				return (EBUSY);
			clrbit(prtabs[maj_d]->pt_enabled.bits, min_d);
		}
		return (0);

#ifdef PRCOMPAT
	/*
	 * PRGETOPRTAB and PRNEXTOPRTAB are used to get the older style
	 * prtab entries for backward compatibility w/ previous releases.
	 */
	case PRGETOPRTAB:
		{
			register struct oprtab *oprtab = (struct oprtab *)data;
			register dev_t bdevno;

			bdevno = oprtab->bmajordev;

			if (!PR_MAJORBDEV_OK(bdevno)) {
				oprtab->bmajordev = NODEV;
			} else {
				mkoprtab(oprtab, prtabs[bdevno]);
			}
		}
		return (0);

	case PRNEXTOPRTAB:
		{
			register struct oprtab *oprtab = (struct oprtab *)data;
			register dev_t bdevno;

			bdevno = oprtab->bmajordev;
			if (bdevno == NODEV)
				bdevno = 0;
			else
				bdevno++;

			while (bdevno < nprbdev) {
				if (PR_MAJORBDEV_OK(bdevno)) {
					mkoprtab(oprtab, prtabs[bdevno]);
					return (0);
				} else {
					bdevno++;
				}
			}
			oprtab->bmajordev = NODEV;
		}
		return (0);
#endif /* PRCOMPAT */

	/*
	 * PRGETUPRTAB and PRNEXTUPRTAB are used to get
	 * the user versions of the prtab entries.
	 */
	case PRGETUPRTAB:
		{
			register struct uprtab *uprtab = (struct uprtab *)data;
			register dev_t bdevno;

			bdevno = uprtab->upt_bmajordev;

			if (!PR_MAJORBDEV_OK(bdevno)) {
				uprtab->upt_bmajordev = NODEV;
			} else {
				mkuprtab(uprtab, prtabs[bdevno]);
			}
		}
		return (0);

	case PRNEXTUPRTAB:
		{
			register struct uprtab *uprtab = (struct uprtab *)data;
			register dev_t bdevno;

			bdevno = uprtab->upt_bmajordev;
			if (bdevno == NODEV)
				bdevno = 0;
			else
				bdevno++;

			while (bdevno < nprbdev) {
				if (PR_MAJORBDEV_OK(bdevno)) {
					mkuprtab(uprtab, prtabs[bdevno]);
					return (0);
				} else {
					bdevno++;
				}
			}
			uprtab->upt_bmajordev = NODEV;
		}
		return (0);

	case PRFLUSH:
		if (prstatus.pr_state != PRDOWN) {
			/* flush w/ no invalidate */
			(void) prflushdev(NODEV, FALSE);
			if (prstatus.pr_state == PRERROR)
				return (EBUSY);
		}
		return (0);

	default:
		return (EINVAL);
	}
}
#endif /* NPR > 0 */
